# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for the ModelModifier class."""

import pathlib
import tracemalloc

from absl.testing import absltest
from absl.testing import parameterized

from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import model_modifier
from ai_edge_quantizer import params_generator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('.')


class ModelModifierTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite'
    )
    self._model_content: bytes = tfl_flatbuffer_utils.get_model_content(
        self._model_path
    )
    self._model = tfl_flatbuffer_utils.read_model(self._model_content)
    self._model_modifier = model_modifier.ModelModifier(self._model)
    self._global_recipe = [
        {
            'regex': '.*',
            'operation': 'FULLY_CONNECTED',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'weight_tensor_config': {
                    'dtype': qtyping.TensorDataType.INT,
                    'num_bits': 8,
                    'symmetric': False,
                    'granularity': qtyping.QuantGranularity.CHANNELWISE,
                    'block_size': 0,
                },
                # Equivalent to WEIGHT_ONLY.
                'compute_precision': qtyping.ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
    ]

  def test_process_constant_map_succeeds(self):
    constant_size = self._model_modifier._process_constant_map(self._model)
    self.assertEqual(constant_size, 202540)

  def test_modify_model_succeeds_with_recipe(self):
    recipe_manager_instance = recipe_manager.RecipeManager()
    params_generator_instance = params_generator.ParamsGenerator(self._model)

    recipe_manager_instance.load_quantization_recipe(self._global_recipe)
    tensor_quantization_params = (
        params_generator_instance.generate_quantization_parameters(
            recipe_manager_instance
        )
    )
    new_model_binary = self._model_modifier.modify_model(
        tensor_quantization_params
    )
    self.assertIsInstance(
        flatbuffer_utils.convert_bytearray_to_object(new_model_binary),
        tfl_flatbuffer_utils.ModelT,
    )
    self.assertLess(len(new_model_binary), len(self._model_content))

  def test_modify_model_preserves_original_model(self):
    recipe_manager_instance = recipe_manager.RecipeManager()
    params_generator_instance = params_generator.ParamsGenerator(self._model)

    recipe_manager_instance.load_quantization_recipe(self._global_recipe)
    tensor_quantization_params = (
        params_generator_instance.generate_quantization_parameters(
            recipe_manager_instance
        )
    )
    model_buffer = flatbuffer_utils.convert_object_to_bytearray(
        self._model_modifier._model
    )
    self._model_modifier.modify_model(tensor_quantization_params)
    self.assertEqual(
        flatbuffer_utils.convert_object_to_bytearray(
            self._model_modifier._model
        ),
        model_buffer,
    )

  def test_modify_model_peak_memory_usage_in_acceptable_range(self):
    """Test ModifyModel peak memory usage."""

    recipe_manager_instance = recipe_manager.RecipeManager()
    params_generator_instance = params_generator.ParamsGenerator(self._model)

    recipe_manager_instance.load_quantization_recipe(self._global_recipe)
    tensor_quantization_params = (
        params_generator_instance.generate_quantization_parameters(
            recipe_manager_instance
        )
    )

    tracemalloc.start()
    self._model_modifier.modify_model(tensor_quantization_params)
    _, mem_peak = tracemalloc.get_traced_memory()

    loosen_mem_use_factor = 4.5
    self.assertLess(mem_peak / len(self._model_content), loosen_mem_use_factor)

  def test_has_transform_before_output_true_dequant(self):
    instructions = {
        'tensor1': qtyping.TensorTransformationInsts(
            'tensor1',
            0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=0,
                    producer=0,
                    consumers=[-1],
                )
            ],
        )
    }
    self.assertTrue(
        self._model_modifier._has_transform_before_output(
            instructions, qtyping.QuantTransformation.ADD_DEQUANTIZE
        )
    )

  def test_has_transform_before_output_false_dequant(self):
    instructions = {
        'tensor1': qtyping.TensorTransformationInsts(
            'tensor1',
            0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_DEQUANTIZE,
                    tensor_id=0,
                    producer=0,
                    consumers=[1],
                )
            ],
        )
    }
    self.assertFalse(
        self._model_modifier._has_transform_before_output(
            instructions, qtyping.QuantTransformation.ADD_DEQUANTIZE
        )
    )

  def test_has_transform_before_output_true_quant(self):
    instructions = {
        'tensor1': qtyping.TensorTransformationInsts(
            'tensor1',
            0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                    tensor_id=0,
                    producer=0,
                    consumers=[-1],
                )
            ],
        )
    }
    self.assertTrue(
        self._model_modifier._has_transform_before_output(
            instructions, qtyping.QuantTransformation.ADD_QUANTIZE
        )
    )

  def test_has_transform_before_output_false_quant(self):
    instructions = {
        'tensor1': qtyping.TensorTransformationInsts(
            'tensor1',
            0,
            instructions=[
                qtyping.TransformationInst(
                    transformation=qtyping.QuantTransformation.ADD_QUANTIZE,
                    tensor_id=0,
                    producer=0,
                    consumers=[1],
                )
            ],
        )
    }
    self.assertFalse(
        self._model_modifier._has_transform_before_output(
            instructions, qtyping.QuantTransformation.ADD_QUANTIZE
        )
    )

  def test_pad_bytearray(self):
    arr = bytearray(b'\x01\x02\x03')
    self._model_modifier._pad_bytearray(arr)
    self.assertLen(arr, 16)
    self.assertEqual(arr, b'\x01\x02\x03' + b'\0' * 13)

    arr = bytearray(b'\x01' * 16)
    self._model_modifier._pad_bytearray(arr)
    self.assertLen(arr, 16)

    arr = bytearray(b'\x01' * 17)
    self._model_modifier._pad_bytearray(arr)
    self.assertLen(arr, 32)


class ModelModifierTestWithSignature(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc.tflite',
    )
    self._model_content: bytes = tfl_flatbuffer_utils.get_model_content(
        self._model_path
    )
    self._model = tfl_flatbuffer_utils.read_model(self._model_content)
    self._model_modifier = model_modifier.ModelModifier(self._model)

  def test_update_signature_defs_succeeds_dequant(self):
    # This is a simplified test that only checks if the function runs without
    # crashing and returns a model. A more thorough test with a model
    # with a known signature was added in `quantizer_test`.
    model_bytearray = flatbuffer_utils.read_model_from_bytearray(
        self._model_content
    )
    updated_model = self._model_modifier._update_signature_defs(
        model_bytearray, bytearray(self._model_content), '_dequant'
    )
    self.assertIsNotNone(updated_model)

  def test_update_signature_defs_succeeds_quant(self):
    # This checks if the function runs without crashing and returns a model.
    model_bytearray = flatbuffer_utils.read_model_from_bytearray(
        self._model_content
    )
    updated_model = self._model_modifier._update_signature_defs(
        model_bytearray, bytearray(self._model_content), '_quantized'
    )
    self.assertIsNotNone(updated_model)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test Hadamard rotation materialization."""

import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import hadamard_rotation
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../../tests/models")
_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig


class HadamardRotationFullyConnectedTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(888)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = None
    self._subgraph = self._test_model.subgraphs[0]
    self._fc_subgraph_op_index = 3
    self._fc_op = self._subgraph.operators[self._fc_subgraph_op_index]
    self._fc_buffer_id = self._subgraph.tensors[self._fc_op.inputs[1]].buffer
    self._op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=self._fc_subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.CHANNELWISE,
            ),
        ),
    )

  def test_materialize_fully_connected_basic(self):
    params = hadamard_rotation.materialize_fully_connected_custom_op(
        self._op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    fc_input = params[0]
    weight = params[1]
    bias = params[2]
    output = params[3]

    self.assertLen(params, 4)
    self.assertIsNone(fc_input.producer)
    self.assertIsNotNone(fc_input.consumers)
    self.assertIsNone(weight.producer)
    self.assertIsNotNone(weight.consumers)
    self.assertIsNone(bias.producer)
    self.assertIsNotNone(bias.consumers)
    self.assertIsNotNone(output.producer)
    self.assertIsNone(output.consumers)
    self.assertEqual(
        fc_input.consumers[0].transformations,
        [qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION],
    )
    self.assertEqual(
        weight.consumers[0].transformations,
        [qtyping.QuantTransformation.QUANTIZE_TENSOR],
    )
    self.assertEqual(
        bias.consumers[0].transformations,
        [qtyping.QuantTransformation.NO_QUANTIZE],
    )
    if output.producer is not None:
      self.assertEqual(
          output.producer.transformations,
          [qtyping.QuantTransformation.NO_QUANTIZE],
      )

  def test_fully_connected_tensorwise_supported(self):
    self._op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=self._fc_subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.TENSORWISE,
            ),
        ),
    )
    params = hadamard_rotation.materialize_fully_connected_custom_op(
        self._op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    self.assertLen(params, 4)
    fc_input = params[0]
    self.assertIsNotNone(fc_input)
    self.assertIsNotNone(fc_input.consumers)
    self.assertIsNotNone(fc_input.consumers[0].parameters)
    self.assertIsInstance(
        fc_input.consumers[0].parameters, qtyping.UniformQuantParams
    )
    if isinstance(
        fc_input.consumers[0].parameters, qtyping.UniformQuantParams
    ):
      self.assertIsNone(fc_input.consumers[0].parameters.quantized_dimension)
    weight = params[1]
    self.assertIsNotNone(weight)
    self.assertIsNotNone(weight.consumers)
    self.assertIsNotNone(weight.consumers[0].parameters)
    self.assertIsInstance(
        weight.consumers[0].parameters, qtyping.UniformQuantParams
    )
    if isinstance(
        weight.consumers[0].parameters, qtyping.UniformQuantParams
    ):
      self.assertIsNone(weight.consumers[0].parameters.quantized_dimension)

  def test_fully_connected_blockwise_supported(self):
    self._op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=self._fc_subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.BLOCKWISE_32,
            ),
        ),
    )
    params = hadamard_rotation.materialize_fully_connected_custom_op(
        self._op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    self.assertLen(params, 4)
    fc_input = params[0]
    self.assertIsNotNone(fc_input)
    self.assertIsNotNone(fc_input.consumers)
    self.assertIsNotNone(fc_input.consumers[0].parameters)
    self.assertIsInstance(
        fc_input.consumers[0].parameters, qtyping.UniformQuantParams
    )
    if isinstance(
        fc_input.consumers[0].parameters, qtyping.UniformQuantParams
    ):
      self.assertEqual(fc_input.consumers[0].parameters.quantized_dimension, 1)
    weight = params[1]
    self.assertIsNotNone(weight)
    self.assertIsNotNone(weight.consumers)
    self.assertIsNotNone(weight.consumers[0].parameters)
    self.assertIsInstance(
        weight.consumers[0].parameters, qtyping.UniformQuantParams
    )
    if isinstance(
        weight.consumers[0].parameters, qtyping.UniformQuantParams
    ):
      self.assertEqual(weight.consumers[0].parameters.quantized_dimension, 1)

  def test_materialize_fully_connected_decomposed(self):
    params = hadamard_rotation.materialize_fully_connected_decomposed(
        self._op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    fc_input = params[0]
    weight = params[1]
    bias = params[2]
    output = params[3]

    self.assertLen(params, 4)
    self.assertEqual(
        fc_input.consumers[0].transformations,
        [qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION],
    )
    self.assertEqual(
        weight.consumers[0].transformations,
        [qtyping.QuantTransformation.QUANTIZE_TENSOR],
    )
    self.assertEqual(
        bias.consumers[0].transformations,
        [qtyping.QuantTransformation.NO_QUANTIZE],
    )
    if output.producer is not None:
      self.assertEqual(
          output.producer.transformations,
          [qtyping.QuantTransformation.NO_QUANTIZE],
      )

  def test_get_tensor_quant_params_basic(self):
    input_tensor = self._subgraph.tensors[self._fc_op.inputs[1]]
    buffer = self._graph_info.buffers[self._fc_buffer_id]
    np_buffer = np.frombuffer(buffer.data, dtype=np.float32).reshape(
        input_tensor.shape
    )
    qparams = hadamard_rotation.get_tensor_quant_params(
        self._op_info,
        self._op_info.op_quant_config.weight_tensor_config,
        np_buffer,
        self._tensor_name_to_qsv,
    )
    self.assertEqual(qparams.num_bits, 8)
    self.assertEqual(qparams.zero_point.all(), 0)
    self.assertEqual(qparams.symmetric, True)
    self.assertIsNotNone(qparams.quantized_data)
    self.assertEqual(qparams.block_size, 0)
    self.assertIsNotNone(qparams.hadamard)
    if qparams.hadamard is not None:
      self.assertEqual(qparams.hadamard.hadamard_size, 32)

  def test_get_tensor_quant_params_golden_1(self):
    test_data = np.ones((6, 6))
    # expected:
    #   [[127   0 127   0 127   0]
    #    [127   0 127   0 127   0]
    #    [127   0 127   0 127   0]
    #    [127   0 127   0 127   0]
    #    [127   0 127   0 127   0]
    #    [127   0 127   0 127   0]]
    expected = np.tile([127, 0], [6, 3])
    qparams = hadamard_rotation.get_tensor_quant_params(
        self._op_info,
        self._op_info.op_quant_config.weight_tensor_config,
        test_data,
        self._tensor_name_to_qsv,
    )
    self.assertIsNotNone(qparams.quantized_data)
    np.testing.assert_array_equal(
        np.array(qparams.quantized_data), expected
    )

  def test_get_tensor_quant_params_golden_2(self):
    # test_data:
    #   [[1 2 1 2 1 2]
    #    [3 4 3 4 3 4]
    #    [1 2 1 2 1 2]
    #    [3 4 3 4 3 4]
    #    [1 2 1 2 1 2]
    #    [3 4 3 4 3 4]]
    test_data = np.tile([[1, 2], [3, 4]], [3, 3])
    # expected:
    #   [[127 -42 127 -42 127 -42]
    #    [127 -18 127 -18 127 -18]
    #    [127 -42 127 -42 127 -42]
    #    [127 -18 127 -18 127 -18]
    #    [127 -42 127 -42 127 -42]
    #    [127 -18 127 -18 127 -18]]
    expected = np.tile([[127, -42], [127, -18]], [3, 3])
    qparams = hadamard_rotation.get_tensor_quant_params(
        self._op_info,
        self._op_info.op_quant_config.weight_tensor_config,
        test_data,
        self._tensor_name_to_qsv,
    )
    self.assertIsNotNone(qparams.quantized_data)
    np.testing.assert_array_equal(
        np.array(qparams.quantized_data), expected
    )

  def test_get_tensor_quant_params_golden_3(self):
    # test_data:
    #   [[[1 2 1 2 1 2]
    #     [3 4 3 4 3 4]
    #     [1 2 1 2 1 2]]
    #    [[3 4 3 4 3 4]
    #     [1 2 1 2 1 2]
    #     [3 4 3 4 3 4]]]
    test_data = np.tile([[1, 2], [3, 4]], [3, 3])
    test_data = np.reshape(test_data, (2, 3, 6))
    # expected:
    #   [[[ 54 -18  54 -18  54 -18]
    #     [127 -18 127 -18 127 -18]
    #     [ 54 -18  54 -18  54 -18]]
    #    [[127 -18 127 -18 127 -18]
    #     [ 54 -18  54 -18  54 -18]
    #     [127 -18 127 -18 127 -18]]]
    expected = np.tile([[54, -18], [127, -18]], [3, 3])
    expected = np.reshape(expected, (2, 3, 6))
    qparams = hadamard_rotation.get_tensor_quant_params(
        self._op_info,
        self._op_info.op_quant_config.weight_tensor_config,
        test_data,
        self._tensor_name_to_qsv,
    )
    self.assertIsNotNone(qparams.quantized_data)
    np.testing.assert_array_equal(
        np.array(qparams.quantized_data), expected
    )

  def test_raise_missing_tensor_content(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: "weight tensor" in str(err)
    ):
      hadamard_rotation.get_tensor_quant_params(
          self._op_info,
          self._op_info.op_quant_config.weight_tensor_config,
          None,
          self._tensor_name_to_qsv,
      )

  def test_raise_qsv_set(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: "static quantization" in str(err)
    ):
      hadamard_rotation.get_tensor_quant_params(
          self._op_info,
          self._op_info.op_quant_config.weight_tensor_config,
          self._graph_info.buffers[self._fc_buffer_id],
          self._graph_info.buffers[self._fc_buffer_id],
      )

  def test_raise_1d_constant(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: "rank >= 2" in str(err)
    ):
      hadamard_rotation.get_tensor_quant_params(
          self._op_info,
          self._op_info.op_quant_config.weight_tensor_config,
          np.array([1.0, 2.0, 3.0]),
          self._tensor_name_to_qsv,
      )


class HadamardRotationEmbeddingLookupTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(888)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "embedding_lookup.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = None

  def test_materialize_embedding_lookup_basic(self):
    subgraph = self._test_model.subgraphs[0]
    embedding_subgraph_op_index = 0
    embedding_op = subgraph.operators[embedding_subgraph_op_index]
    op_info = qtyping.OpInfo(
        op=embedding_op,
        op_name=_TFLOpName.EMBEDDING_LOOKUP,
        subgraph_op_index=embedding_subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.CHANNELWISE,
            ),
        ),
    )
    params = hadamard_rotation.materialize_embedding_lookup_custom_op(
        op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    self.assertLen(params, 3)
    lookup = params[0]
    value = params[1]
    output = params[2]
    self.assertIsNone(lookup.producer)
    self.assertIsNotNone(lookup.consumers)
    self.assertIsNone(value.producer)
    self.assertIsNotNone(value.consumers)
    self.assertIsNotNone(output.producer)
    self.assertIsNone(output.consumers)
    self.assertEqual(
        lookup.consumers[0].transformations,
        [qtyping.QuantTransformation.NO_QUANTIZE],
    )
    self.assertEqual(
        value.consumers[0].transformations,
        [qtyping.QuantTransformation.QUANTIZE_TENSOR],
    )
    if output.producer is not None:
      self.assertEqual(
          output.producer.transformations,
          [qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION],
      )

  def test_materialize_embedding_lookup_decomposed(self):
    subgraph = self._test_model.subgraphs[0]
    embedding_subgraph_op_index = 0
    embedding_op = subgraph.operators[embedding_subgraph_op_index]
    op_info = qtyping.OpInfo(
        op=embedding_op,
        op_name=_TFLOpName.EMBEDDING_LOOKUP,
        subgraph_op_index=embedding_subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=True,
                granularity=qtyping.QuantGranularity.CHANNELWISE,
            ),
        ),
    )
    params = hadamard_rotation.materialize_embedding_lookup_decomposed(
        op_info,
        self._graph_info,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
        tensor_name_to_qsv=self._tensor_name_to_qsv,
    )
    self.assertLen(params, 3)
    lookup = params[0]
    value = params[1]
    output = params[2]
    self.assertEqual(
        lookup.consumers[0].transformations,
        [qtyping.QuantTransformation.NO_QUANTIZE],
    )
    self.assertEqual(
        value.consumers[0].transformations,
        [qtyping.QuantTransformation.QUANTIZE_TENSOR],
    )
    if output.producer is not None:
      self.assertEqual(
          output.producer.transformations,
          [qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION],
      )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
from typing import cast

from absl.testing import absltest
from absl.testing import parameterized
import ml_dtypes
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile("../../tests/models")
_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig


class NaiveMinMaxQuantizeTest(parameterized.TestCase):
  """Tests for general functions innaive min-max quantize algorithm.

  See naive_min_max_quantize_op_tests for op specific tests.
  """

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "conv_fc_mnist.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = {}

  @parameterized.parameters(
      (qtyping.QuantGranularity.TENSORWISE),
      (qtyping.QuantGranularity.CHANNELWISE),
  )
  def test_init_qsvs(self, granularity):
    # Read from Model Explorer.
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_op_index = 3
    fc_op = subgraph0.operators[subgraph_op_index]
    op_info = qtyping.OpInfo(
        op=fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=_TensorQuantConfig(
                8,
                symmetric=True,
                granularity=granularity,
            ),
        ),
    )

    initial_qsvs = naive_min_max_quantize.init_qsvs(
        op_info,
        self._graph_info,
    )
    self.assertIn("sequential/flatten/Reshape", initial_qsvs)
    input_tensor_qsv = initial_qsvs["sequential/flatten/Reshape"]
    self.assertEmpty(input_tensor_qsv)
    self.assertIn(
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd",
        initial_qsvs,
    )
    output_tensor_qsv = initial_qsvs[
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd"
    ]
    self.assertEmpty(output_tensor_qsv)

    self.assertIn("arith.constant1", initial_qsvs)
    weight_tensor_qsv = initial_qsvs["arith.constant1"]
    if granularity is qtyping.QuantGranularity.CHANNELWISE:
      mins_maxs_shape = (32, 1)
    else:
      mins_maxs_shape = (1, 1)
    self.assertTupleEqual(weight_tensor_qsv["min"].shape, mins_maxs_shape)
    self.assertTupleEqual(weight_tensor_qsv["max"].shape, mins_maxs_shape)

    self.assertIn("arith.constant2", initial_qsvs)
    bias_tensor_qsv = initial_qsvs["arith.constant2"]
    if granularity is qtyping.QuantGranularity.CHANNELWISE:
      mins_maxs_shape = (32,)
    else:
      mins_maxs_shape = (1,)
    self.assertTupleEqual(bias_tensor_qsv["min"].shape, mins_maxs_shape)
    self.assertTupleEqual(bias_tensor_qsv["max"].shape, mins_maxs_shape)

    initial_qsvs = naive_min_max_quantize.init_qsvs(
        op_info,
        self._graph_info,
        inputs_to_ignore=[0],
        outputs_to_ignore=[0],
    )
    self.assertNotIn("sequential/flatten/Reshape", initial_qsvs)
    self.assertNotIn(
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd",
        initial_qsvs,
    )

  def test_min_max_calibrate(self):
    # Sample input/output data for the fc op.
    tensor_content_map = {
        "sequential/flatten/Reshape": np.array([[1, 2, 3, 4, 5]]),
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd": np.array(
            [[6, 7, 8, 9, 10]]
        ),
    }
    # Read from Model Explorer.
    subgraph0 = self._test_model.subgraphs[0]
    fc_op = subgraph0.operators[3]
    # ignore 1(weight), and 2(bias) in inputs.
    op_qsvs = naive_min_max_quantize.min_max_calibrate(
        fc_op,
        self._graph_info,
        tensor_content_map,
        [1, 2],
        [],
    )
    self.assertIn("sequential/flatten/Reshape", op_qsvs)
    input_tensor_qsv = op_qsvs["sequential/flatten/Reshape"]
    self.assertTupleEqual(input_tensor_qsv["min"].shape, (1, 1))
    self.assertEqual(input_tensor_qsv["min"], np.array([[1]]))
    self.assertTupleEqual(input_tensor_qsv["max"].shape, (1, 1))
    self.assertEqual(input_tensor_qsv["max"], np.array([[5]]))
    self.assertIn(
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd",
        op_qsvs,
    )
    output_tensor_qsv = op_qsvs[
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd"
    ]
    self.assertTupleEqual(output_tensor_qsv["min"].shape, (1, 1))
    self.assertEqual(output_tensor_qsv["min"], np.array([[6]]))
    self.assertTupleEqual(output_tensor_qsv["max"].shape, (1, 1))
    self.assertEqual(output_tensor_qsv["max"], np.array([[10]]))
    # weight and bias are excluded.
    self.assertNotIn("arith.constant1", op_qsvs)
    self.assertNotIn("arith.constant2", op_qsvs)

  def test_get_tensor_quant_params_for_blockwise_weight(self):
    subgraph0 = self._test_model.subgraphs[0]
    subgraph_op_index = 3
    fc_op = subgraph0.operators[subgraph_op_index]
    weight_tensor_config = _TensorQuantConfig(
        num_bits=4,
        symmetric=True,
        granularity=qtyping.QuantGranularity.BLOCKWISE_32,
    )
    op_info = qtyping.OpInfo(
        op=fc_op,
        op_name=_TFLOpName.FULLY_CONNECTED,
        subgraph_op_index=subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=weight_tensor_config,
        ),
    )
    test_data = np.random.uniform(low=-10, high=10, size=(4, 32)).astype(
        np.float32
    )
    quant_params = naive_min_max_quantize.get_tensor_quant_params(
        op_info=op_info,
        tensor_quant_config=weight_tensor_config,
        tensor_content=test_data,
    )
    zp = quant_params.zero_point
    self.assertEqual(zp.shape, (4, 1))
    self.assertTrue(np.array_equal(zp, np.zeros([4, 1])))

    self.assertEqual(quant_params.scale.shape, (4, 1))
    expected_scales = np.max(np.abs(test_data), axis=1, keepdims=True) / 7.0
    expected_scales = (
        expected_scales.astype(ml_dtypes.bfloat16)
        .astype(np.float16)
        .astype(np.float32)
    )
    self.assertTrue(np.allclose(quant_params.scale, expected_scales, atol=1e-5))

    self.assertIsNotNone(quant_params.quantized_data)
    self.assertTupleEqual(
        cast(np.ndarray, quant_params.quantized_data).shape, test_data.shape
    )
    self.assertEqual(quant_params.block_size, 32)
    self.assertEqual(quant_params.quantized_dimension, 1)

  def test_calibrate_ignores_inf_min_max(self):
    """Tests that calibration ignores infinity values."""
    # Sample input/output data for the fc op.
    input_tensor_name = "sequential/flatten/Reshape"
    output_tensor_name = (
        "sequential/dense/MatMul;sequential/dense/Relu;sequential/dense/BiasAdd"
    )
    bloat16_inf = 3.39e38
    tensor_content_map = {
        input_tensor_name: np.array(
            [[-np.inf, 1.0, 5.0, np.inf, bloat16_inf]], dtype=np.float32
        ),
        output_tensor_name: np.array(
            [[6.0, 7.0, -bloat16_inf, 9.0, np.inf]], dtype=np.float32
        ),
    }
    # Read from Model Explorer.
    subgraph0 = self._test_model.subgraphs[0]
    fc_op = subgraph0.operators[3]
    op_qsvs = naive_min_max_quantize.min_max_calibrate(
        fc_op,
        self._graph_info,
        tensor_content_map,
        inputs_to_ignore=[1, 2],  # Ignore weight and bias.
        outputs_to_ignore=[],
    )

    self.assertIn(input_tensor_name, op_qsvs)
    self.assertEqual(op_qsvs[input_tensor_name]["min"], 1.0)
    self.assertEqual(op_qsvs[input_tensor_name]["max"], 5.0)

    self.assertIn(output_tensor_name, op_qsvs)
    self.assertEqual(op_qsvs[output_tensor_name]["min"], 6.0)
    self.assertEqual(op_qsvs[output_tensor_name]["max"], 9.0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Hadamard rotation pattern transformation."""

from flatbuffers import flexbuffers
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


def _to_flexbuffer(
    hadamard_size: int,
    random_binary_vector: list[np.int8],
) -> bytes:
  """Converts hadamard_size to flexbuffer."""
  fbb = flexbuffers.Builder()
  with fbb.Map():
    fbb.Int('hadamard_size', hadamard_size)
    fbb.VectorFromElements('random_binary_vector', random_binary_vector)
  return fbb.Finish()


def _update_embedding_lookup_consumers(
    transformation: transformation_utils.TransformationInput,
    new_tensor_id: int,
) -> bool:
  """Updates the consumers of the embedding lookup op to use the new tensor.

  Args:
    transformation: The transformation input to update the consumers of.
    new_tensor_id: The new tensor id to use as the input to the embedding lookup
      consumers.
  """
  for consumer in transformation.consumers:
    # If the consumer is a graph output and not an op, we can ignore it here
    # since the graph output will be updated later.
    if consumer == -1:
      continue
    consumer_op = transformation.subgraph.operators[consumer]
    # Find the input that was attached to the insertion point, and replace it
    # with the new tensor.
    for i in range(len(consumer_op.inputs)):
      if consumer_op.inputs[i] == transformation.tensor_id:
        consumer_op.inputs[i] = new_tensor_id


def _update_fully_connected_consumers(
    transformation: transformation_utils.TransformationInput,
    new_tensor_id: int,
) -> bool:
  """Updates the fully connected op(s) to use the new tensor.

  Since the new tensor is inserted to the fully_connected's input, we need to
  scan each consumer (in case of multiple fully_connected ops), and update
  the input tensor to the new tensor.

  Args:
    transformation: The transformation input to update the consumers of.
    new_tensor_id: The new tensor id to use as the input to the fully connected
      consumers.

  Returns:
    True if the fully connected op(s) were updated to use the new tensor.
  """
  updated = False
  for consumer in transformation.consumers:
    if (
        transformation_utils.get_schema_op_id(transformation, consumer)
        == schema_py_generated.BuiltinOperator.FULLY_CONNECTED
    ):
      transformation.subgraph.operators[consumer].inputs[0] = new_tensor_id
      updated = True
  return updated


def insert_hadamard_rotation(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Inserts a custom aeq.hadamard_rotation op on this tensor.

  This function works for float32 tensors only.

  Args:
    transformation_input: The transformation input to insert the custom op on.

  Returns:
    The transformation info of the inserted custom op.

  Raises:
    ValueError: If the transformation input is not a uniform quantization
    transformation.
    ValueError: If the Hadamard quantization params are not set.
    ValueError: If the tensor is not a float32 tensor.
    ValueError: If no supported ops were found as the tensor's producer or
    consumers.
  """
  if not isinstance(
      transformation_input.quant_params, qtyping.UniformQuantParams
  ):
    raise ValueError('Hadamard rotation supports uniform quantization only')

  if transformation_input.quant_params.hadamard is None:
    raise ValueError(
        'Hadamard rotation quantization params are not set but op insertion is'
        ' requested.'
    )

  tensor = transformation_input.subgraph.tensors[transformation_input.tensor_id]
  if tensor.type != schema_py_generated.TensorType.FLOAT32:
    raise ValueError(
        'The Hadamard rotation op supports float32 tensors only. Got'
        f' {tensor.type} tensor.'
    )

  # Create new custom op with the current tensor as input and a new activation
  # tensor as output.
  custom_op_code_idx = transformation_utils.add_op_code(
      schema_py_generated.BuiltinOperator.CUSTOM,
      transformation_input.op_codes,
      'aeq.hadamard_rotation',
  )
  custom_op = schema_py_generated.OperatorT()
  custom_op.opcodeIndex = custom_op_code_idx
  custom_op.inputs = [transformation_input.tensor_id]
  custom_op.customOptions = _to_flexbuffer(
      transformation_input.quant_params.hadamard.hadamard_size,
      transformation_input.quant_params.hadamard.random_binary_vector.tolist(),
  )
  new_tensor_id = transformation_utils.add_new_activation_tensor(
      tensor.name + b'_rotated',
      tensor.shapeSignature
      if tensor.shapeSignature is not None
      else tensor.shape,
      schema_py_generated.TensorType.FLOAT32,
      transformation_input.subgraph,
  )
  custom_op.outputs = [new_tensor_id]

  # Update the users of this tensor to use the new tensor.
  if (
      transformation_utils.get_producer_schema_op_id(transformation_input)
      == schema_py_generated.BuiltinOperator.EMBEDDING_LOOKUP
  ):
    _update_embedding_lookup_consumers(transformation_input, new_tensor_id)
  elif not _update_fully_connected_consumers(
      transformation_input, new_tensor_id
  ):
    raise ValueError(
        'The Hadamard rotation op supports embedding lookup and fully connected'
        ' ops only, but no such ops were found.'
    )

  # If the tensor is a graph output, we need to replace the tensor with the
  # new tensor.
  for i, output in enumerate(transformation_input.subgraph.outputs):
    if output == transformation_input.tensor_id:
      transformation_input.subgraph.outputs[i] = new_tensor_id

  # Find the actual insertion point. The insertion point should be after the
  # producer op and before the first consumer op. The max() operation ensures
  # that we're not using -1 as the insertion point.
  first_consumer_id = min(transformation_input.consumers)
  op_id = max(transformation_input.producer + 1, first_consumer_id)

  # Insert the custom op.
  transformation_input.subgraph.operators.insert(op_id, custom_op)

  return qtyping.TransformationInfo(
      op_id=op_id,
      num_ops_added=1,
      output_tensor_id=new_tensor_id,
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import octav
from ai_edge_quantizer.algorithms.uniform_quantize.op_architecture_tests import test_utils as op_test_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(
    "../../../tests/models"
)


class ReluTest(op_test_utils.BaseQuantizeTest):

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "single_relu.tflite"
    )
    self._op_test_info = op_test_utils.OpTestInfo(
        test_model=tfl_flatbuffer_utils.read_model(self._test_model_path),
        op_tensor_names={},
        input_range=(np.array([[-10]]), np.array([[10]])),
        output_range=(np.array([[0]]), np.array([[10]])),
    )
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._op_test_info.test_model.subgraphs[0].tensors,
        buffers=self._op_test_info.test_model.buffers,
    )

  @parameterized.parameters(
      # get_tensor_quant_params_func, activations_num_bits, symmetric
      (naive_min_max_quantize.get_tensor_quant_params, 8, True),
      (naive_min_max_quantize.get_tensor_quant_params, 8, False),
      (naive_min_max_quantize.get_tensor_quant_params, 16, True),
      (octav.get_tensor_quant_params, 8, True),
      (octav.get_tensor_quant_params, 16, True),
  )
  def test_materialize_relu_succeeds(
      self, get_tensor_quant_params_func, activations_num_bits, symmetric
  ):
    activation_config = test_utils.get_static_activation_quant_setting(
        activations_num_bits, symmetric
    )
    op_quant_config = test_utils.get_static_op_quant_config(activation_config)

    # Read from Model Explorer.
    subgraph0 = self._op_test_info.test_model.subgraphs[0]
    subgraph_op_id = 0
    op = subgraph0.operators[subgraph_op_id]
    op_info = qtyping.OpInfo(
        op=op,
        op_name=qtyping.TFLOperationName.RELU,
        subgraph_op_index=subgraph_op_id,
        op_quant_config=op_quant_config,
    )

    # Test settings.
    op_tensor_names = {}
    op_tensor_names["input"] = "serving_default_input:0"
    op_tensor_names["output"] = "PartitionedCall:0"
    self._op_test_info.op_tensor_names = op_tensor_names
    self._test_no_weights_op(
        op_info,
        self._graph_info,
        self._op_test_info,
        common_quantize.materialize_relu,
        get_tensor_quant_params_func,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A library of comparator function to be used by validation function."""

from collections.abc import Callable
from typing import Any, Tuple
import numpy as np


def get_validation_func(
    func_name: str,
) -> Callable[[np._typing.ArrayLike, np._typing.ArrayLike], Any]:
  """Returns a validation function based on the function name.

  Args:
    func_name: name of the validation function

  Returns:
    a validation function

  Raises:
    ValueError: if the function name is not supported
  """
  if func_name == "mse":
    return mean_squared_difference
  elif func_name == "median_diff_ratio":
    return median_diff_ratio
  elif func_name == "cosine_similarity":
    return cosine_similarity
  elif func_name == "kl_divergence":
    return kl_divergence
  elif func_name == "snr":
    return signal_to_noise_ratio
  else:
    raise ValueError(f"Validation function {func_name} not supported")


def mean_squared_difference(
    data1: np._typing.ArrayLike, data2: np._typing.ArrayLike
) -> float:
  """Calculates the mean squared difference between data1 & data2.

  ref: https://en.wikipedia.org/wiki/Mean_squared_error

  Args:
    data1: input data to be used for comparison
    data2: input data to be used for comparison, data1 & 2 must be of the same
      shape

  Returns:
    a float value representing the MSD between data1 & 2

  Raises:
    ValueError: if the two inputs don't have the same number of elements
  """
  data1, data2 = _preprocess_same_size_arrays(data1, data2)
  # special handling for tensor of size 0
  if data1.size == 0:
    return float(0)
  return float(np.square(np.subtract(data1, data2)).mean())


def median_diff_ratio(
    data1: np._typing.ArrayLike,
    data2: np._typing.ArrayLike,
    tolerance_threshold=1e-6,
) -> float:
  """Calculates the median absolute diff ratio between data1 & data2.

  mdr = median(abs(data1 - data2) / data2)

  Args:
    data1: input data to be used for comparison
    data2: input data to be used for comparison, data1 & 2 must be of the same
      shape
    tolerance_threshold: a float value to be used as a threshold to avoid
      division by zero

  Returns:
    a float value representing the median diff ratio between data1 & 2

  Raises:
    ValueError: if the two inputs don't have the same number of elements
  """
  data1, data2 = _preprocess_same_size_arrays(data1, data2)
  # special handling for tensor of size 0
  if data1.size == 0:
    return float(0)
  diff = abs(data1 - data2)
  demoninator = abs(data2) + tolerance_threshold
  median_ratio = np.median(diff / demoninator)
  return median_ratio


def cosine_similarity(
    data1: np._typing.ArrayLike,
    data2: np._typing.ArrayLike,
) -> float:
  """Calculates the cosine similarity between data1 & data2.

  ref: https://en.wikipedia.org/wiki/Cosine_similarity

  Args:
    data1: input data to be used for comparison
    data2: input data to be used for comparison, data1 & 2 must be of the same
      shape

  Returns:
    a float value representing the cosine similarity between data1 & 2

  Raises:
    ValueError: if the two inputs don't have the same number of elements
  """
  data1, data2 = _preprocess_same_size_arrays(data1, data2)
  # special handling for tensor of size 0
  if data1.size == 0:
    return float(0)
  norm_data1 = np.linalg.norm(data1)
  norm_data2 = np.linalg.norm(data2)
  # special handling for tensor of length 0
  if norm_data1 == 0 and norm_data2 == 0:
    return 1.0
  if norm_data1 == 0 or norm_data2 == 0:
    return 0.0
  return np.dot(data1, data2) / (norm_data1 * norm_data2)


def kl_divergence(
    data1: np._typing.ArrayLike,
    data2: np._typing.ArrayLike,
    epsilon: float = 1e-9,
) -> float:
  """Calculates the KL divergence between data1 & data2.

  KL(data2 || data1) = sum(data2 * log(data2 / data1)).
  data2 is treated as the true distribution P, and data1 as the
  approximated distribution Q.
  Non-positive values in data1 and data2 are clipped to 0 before
  KL divergence calculation. Epsilon is added to avoid log(0) and
  division by zero.

  Args:
    data1: input data to be used for comparison (distribution Q)
    data2: input data to be used for comparison (distribution P), data1 & 2 must
      be of the same shape
    epsilon: small value to avoid log(0) and division by zero.

  Returns:
    A float value representing the KL divergence between data1 & 2.

  Raises:
    ValueError: if the two inputs don't have the same number of elements.
  """
  data1, data2 = _preprocess_same_size_arrays(data1, data2)
  # special handling for tensor of size 0
  if data1.size == 0:
    return float(0)

  p = np.maximum(0, data2)
  q = np.maximum(0, data1)

  return float(np.sum(p * np.log((p + epsilon) / (q + epsilon))))


def signal_to_noise_ratio(
    noisy_signal: np._typing.ArrayLike,
    signal: np._typing.ArrayLike,
    epsilon: float = 1e-9,
) -> float:
  """Calculates the signal to noise ratio between noisy_signal & signal.

  SNR = P_signal / P_noise, where signal is treated as the clean signal and
  noisy_signal-signal is treated as the noise samples.
  P_signal = mean(signal^2)
  P_noise = mean((noisy_signal-signal)^2) = mse(noisy_signal, signal)

  Args:
    noisy_signal: Input data to be used for comparison (e.g. noisy signal).
    signal: Input data to be used for comparison (e.g. clean signal),
      noisy_signal & signal must be of the same shape.
    epsilon: Small value to avoid division by zero.

  Returns:
    A float value representing the SNR between noisy_signal & signal.

  Raises:
    ValueError: If the two inputs don't have the same number of elements.
  """
  noisy_signal, signal = _preprocess_same_size_arrays(noisy_signal, signal)
  if signal.size == 0:
    return float(0)

  mse = mean_squared_difference(noisy_signal, signal)
  signal_power = float(np.square(signal).mean())
  snr = signal_power / (mse + epsilon)
  return snr


def _preprocess_same_size_arrays(
    data1: np._typing.ArrayLike, data2: np._typing.ArrayLike
) -> Tuple[np.ndarray, np.ndarray]:
  """Flattens and removes the nan, inf, and -inf values from the input data.

  Args:
    data1: input data to be used for comparison
    data2: input data to be used for comparison, data1 & 2 must be of the same
      shape

  Returns:
    a tuple of the preprocessed data1 & 2

  Raises:
    ValueError: if the two inputs don't have the same number of elements
  """
  data1 = np.array(data1, dtype=np.float32).flatten()
  data2 = np.array(data2, dtype=np.float32).flatten()
  if np.shape(data1) != np.shape(data2):
    raise ValueError("data1 & data2 must be of the same size")
  data1 = np.nan_to_num(data1, nan=1e-9, neginf=-1e9, posinf=1e9)
  data2 = np.nan_to_num(data2, nan=1e-9, neginf=-1e9, posinf=1e9)

  return data1, data2



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities for model calibration."""

import copy
import json
from typing import Any, Union

import numpy as np

import os
import io
from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import constrained_ops_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_SignatureInput = dict[str, Any]
_OpQuantConstraint = common_utils.OpQuantConstraint
_SignatureData = dict[
    str, list[str]
]  # signature_key -> list of signature_names.


class NumpyEncoder(json.JSONEncoder):
  """JSON Encoder for Numpy types."""

  def default(self, o):
    if isinstance(o, np.integer):
      return int(o)
    elif isinstance(o, np.floating):
      return float(o)
    elif isinstance(o, np.ndarray):
      return o.tolist()
    return super().default(o)


def _update_moving_average(
    smoothing_factor: Union[np.ndarray, float],
    w: np.ndarray,
    update: np.ndarray,
) -> np.ndarray:
  """Updates weight w with moving average.

  Args:
    smoothing_factor: Smoothing factor used to update w.
    w: Weights to be updated.
    update: Value used for update.

  Returns:
    Weighted sum of w and update.
  """
  return smoothing_factor * w + (1.0 - smoothing_factor) * update


def moving_average_update(
    qsv: qtyping.QSV, new_qsv: qtyping.QSV, smoothing_factor: float = 0.95
) -> qtyping.QSV:
  """Update the QSV (i.e., min/max) using moving average.

  Args:
    qsv: The quantization statistical value of the tensor (min/max) that need to
      be updated.
    new_qsv: The new QSVs (e.g., from new round of calibration).
    smoothing_factor: The weight of moving average.

  Returns:
    The updated QSV for the tensor.
  """
  if not qsv:
    return new_qsv

  updated_qsv = {}
  updated_qsv["min"] = _update_moving_average(
      smoothing_factor, qsv["min"], new_qsv["min"]
  )

  updated_qsv["max"] = _update_moving_average(
      smoothing_factor, qsv["max"], new_qsv["max"]
  )
  return updated_qsv


def min_max_update(qsv: qtyping.QSV, new_qsv: qtyping.QSV) -> qtyping.QSV:
  """Update the QSV with minimum min values and maximum max values.

  Args:
    qsv: The quantization statistical value of the tensor (min/max) that need to
      be updated.
    new_qsv: The new QSVs (e.g., from new round of calibration).

  Returns:
    The updated QSV for the tensor.
  """
  if not qsv:
    return new_qsv

  updated_qsv = {}
  updated_qsv["min"] = np.minimum(qsv["min"], new_qsv["min"])
  updated_qsv["max"] = np.maximum(qsv["max"], new_qsv["max"])
  return updated_qsv


def load_calibration_results(
    file_path: str,
) -> tuple[dict[str, qtyping.QSV], dict[str, Any]]:
  """Loads calibration results from a file.

  Args:
    file_path: Path to the calibration results file.

  Returns:
    A tuple of (model_qsvs, metadata).
  """
  with open(file_path) as json_file:
    data = json.load(json_file)

  metadata = {}
  if "model_qsvs" in data:
    model_qsvs = data["model_qsvs"]
    metadata = data.get("metadata", {})
  else:
    model_qsvs = data

  # Convert lists back to numpy arrays
  for _, qsv in model_qsvs.items():
    if "min" in qsv:
      qsv["min"] = np.array(qsv["min"])
    if "max" in qsv:
      qsv["max"] = np.array(qsv["max"])

  return model_qsvs, metadata


def _find_overall_min_max(
    qsv: qtyping.QSV, tensor_names: list[str]
) -> tuple[np.ndarray, np.ndarray]:
  """Finds the overall minimum and maximum values for the given tensors.

  Args:
    qsv: The quantization statistical value of the tensor (min/max).
    tensor_names: The list of tensor names to find the minimum and maximum
      values.

  Returns:
    The minimum and maximum values for the given tensors.
  """
  min_value = np.inf
  max_value = -np.inf
  for tensor_name in tensor_names:
    min_value = min(min_value, qsv[tensor_name]["min"])
    max_value = max(max_value, qsv[tensor_name]["max"])
  return min_value, max_value


class CalibrationQsvAlignmentUtils:
  """Calibration utils for alignment of QSVs.

  This class is used to align QSVs for a given model. It builds a list of ops
  that need to be constrained to the same scale as the input. Based on this
  list, it finds the corresponding tensor names for a given signature data.
  """

  def __init__(self, model_path: str):
    self._same_as_input_scale_ops = (
        constrained_ops_utils.get_constrained_op_list(
            _OpQuantConstraint.SAME_AS_INPUT_SCALE
        )
    )

    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(model_path)
    self._flatbuffer_object = tfl_flatbuffer_utils.read_model(model_path)

    signature_keys = list(tfl_interpreter.get_signature_list().keys())

    # Build a dict of signature runners.
    self._signature_runners = {}
    for signature_key in signature_keys:
      signature_runner = tfl_interpreter.get_signature_runner(signature_key)
      self._signature_runners[signature_key] = signature_runner

  def _search_tensor_by_signature_name(
      self, signature_key: str, signature_input_output_name: str, verbose=False
  ) -> list[str]:
    """Searches for a tensor name for a given signature by signature input or output name.

    Args:
      signature_key: Name of the signature.
      signature_input_output_name: Name of the input or output in the signature.
      verbose: Flag to enable verbose output.

    Returns:
      The list with one or two tensor names. The first one is the input tensor
      name, and the second one is the output tensor name.
    """

    if verbose:
      print("Searching tensor by signature name.")

    tensor_names = []

    # Search among inputs.
    input_details = self._signature_runners[signature_key].get_input_details()
    if signature_input_output_name in input_details.keys():
      tensor_names.append(input_details[signature_input_output_name]["name"])

    # Search among outputs.
    output_details = self._signature_runners[signature_key].get_output_details()
    if signature_input_output_name not in output_details:
      if not tensor_names:
        raise ValueError(
            f"Signature {signature_key} does not have input or output"
            f" `{signature_input_output_name}`"
        )
      return tensor_names

    output_tensor_name = output_details[signature_input_output_name]["name"]
    if verbose:
      print(
          ">> Starting recursive search for the output tensor name:"
          f" {output_tensor_name}"
      )

    idx = self._signature_runners[signature_key]._subgraph_index  # pylint: disable=protected-access
    subgraph = self._flatbuffer_object.subgraphs[idx]
    graph_info = qtyping.GraphInfo(
        subgraph.tensors, self._flatbuffer_object.buffers
    )

    # Recursively search the graph for the output tensor name since it may be
    # `SAME_AS_INPUT` constrainted.
    operators = copy.deepcopy(subgraph.operators)
    tensor_name = self._search_reverse_order_recursively(
        graph_info, operators, output_tensor_name, indent="  ", verbose=verbose
    )
    tensor_names.append(tensor_name)

    if verbose:
      print(f"\n\nFound tensor name: {tensor_name}")

    return tensor_names

  def _search_reverse_order_recursively(
      self,
      graph_info: qtyping.GraphInfo,
      operators: list[Any],
      output_tensor_name: str,
      indent: str,
      verbose: bool = False,
  ):
    """Searches for a tensor name in reverse order recursively.

    Stop criteria is when the tensor belongs to an operator that is not
    `SAME_AS_INPUT` constrainted.

    Args:
      graph_info: Graph information.
      operators: List of operators.
      output_tensor_name: Name of the output tensor to search for.
      indent: Indentation string for debug output.
      verbose: Flag to enable verbose output.

    Returns:
      The name of the tensor found, or None if not found.
    """
    op_codes = self._flatbuffer_object.operatorCodes

    while operators:
      op = operators.pop()
      op_code = op_codes[op.opcodeIndex].builtinCode
      op_name = flatbuffer_utils.opcode_to_name(
          self._flatbuffer_object, op.opcodeIndex
      )
      if op_code not in tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME:
        continue
      for output_idx in op.outputs:
        if output_tensor_name == tfl_flatbuffer_utils.get_tensor_name(
            graph_info.subgraph_tensors[output_idx]
        ):
          dbg_str = (
              f"{indent}>> Found `{op_name}`, output tensor"
              f" '{output_tensor_name}'"
          )

          if op_name not in self._same_as_input_scale_ops:
            if verbose:
              print(f"{dbg_str}, returning...")
            return output_tensor_name

          if verbose:
            print(f"{dbg_str}, with SAME_AS_INPUT, search recursively among:")
          for input_idx in op.inputs:
            input_tensor_name = graph_info.subgraph_tensors[
                input_idx
            ].name.decode("utf-8")

            if verbose:
              print(f"{indent}    Input: {input_tensor_name}")

            return self._search_reverse_order_recursively(
                graph_info,
                operators,
                input_tensor_name,
                indent=f"{indent}  ",
                verbose=verbose,
            )
    return output_tensor_name

  def align_quant_stats(
      self, qsv: dict[str, Any], signature_data: _SignatureData
  ) -> tuple[np.ndarray, np.ndarray]:
    """Aligns quantization statistics for a given signature data.

    This function takes quantization statistics and signature data as input,
    identifies the tensors associated with the signature data, and aligns
    the quantization statistics of these tensors by setting their minimum
    and maximum values to the same value. This ensures that the tensors
    have the same quantization parameters.

    Args:
      qsv: Quantization statistics.
      signature_data: Signature data.

    Returns:
      Tuple of min and max values.
    """
    # Go over all signature info and find the corresponding tensor names.
    tensor_names = []
    for signature_key, signature_names in signature_data.items():
      for signature_name in signature_names:
        tensor_name = self._search_tensor_by_signature_name(
            signature_key, signature_name
        )
        tensor_names.extend(tensor_name)

    # Find min and max values accross all tensors.
    min_value, max_value = _find_overall_min_max(qsv, tensor_names)

    # Overwrite the min and max values in the QSV.
    for tensor_name in tensor_names:
      qsv[tensor_name]["min"] = min_value
      qsv[tensor_name]["max"] = max_value

    return min_value, max_value

  def update_quant_stats(
      self,
      qsv: dict[str, Any],
      signature_data: _SignatureData,
      min_value: np.ndarray,
      max_value: np.ndarray,
  ):
    """Updates quantization statistics for a given signature data.

    This function updates the quantization statistics with the provided min, max
    values for the tensors specified in the signature data.

    Args:
      qsv: Quantization statistics.
      signature_data: Signature data.
      min_value: Minimum value to update.
      max_value: Maximum value to update.

    Returns:
      Updated quantization statistics.
    """
    # Go over all signature info and find the corresponding tensor names.
    tensor_names = []
    for signature_key, signature_names in signature_data.items():
      for signature_name in signature_names:
        tensor_name = self._search_tensor_by_signature_name(
            signature_key, signature_name
        )
        tensor_names.extend(tensor_name)

    # Overwrite the min and max values in the QSV.
    for tensor_name in tensor_names:
      qsv[tensor_name]["min"] = min_value
      qsv[tensor_name]["max"] = max_value



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implements the OCTAV quantization."""

import dataclasses
from typing import Any, Optional, Sequence, Union
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils

ALGORITHM_KEY = "OCTAV"


def _guess_clipping_with_octav(
    x: np.ndarray,
    bits: int,
    axis: Union[int, Sequence[int]],
    max_iterations: int,
    exponent_divisor: float,
    early_stop: bool = True,
) -> np.ndarray:
  """Returns a tensor of absolute clipping constants for a tensor using OCTAV.

  This method implements equation (6) from the OCTAV paper:
  https://arxiv.org/abs/2206.06501

  Args:
    x: Tensor data to return guesses for.
    bits: Number of bits used during quantization.
    axis: Axis to reduce the tensor along to get the guesses.
    max_iterations: Number of Newton-Raphson iterations to use.
    exponent_divisor: What factor to divide the 4^-bits term by. In the paper,
      3.0 is optimal for signed ints and 12.0 for unsigned ints.
    early_stop: If True, stop the iteration if the guess doesn't change.

  Returns:
    A tensor of shape [num_channels] with clipping constant guesses.
  """
  magnitude = np.abs(x)
  x_reduced = np.mean(x, axis=axis, keepdims=True)
  old_guess = np.zeros(x_reduced.shape)
  guess = np.ones(x_reduced.shape)
  for _ in range(max_iterations):
    if early_stop and np.allclose(guess, old_guess):
      break
    guess_broadcasted = np.broadcast_to(guess, magnitude.shape)
    guess_mask = np.asarray(magnitude < guess_broadcasted, dtype=x.dtype)
    numerator = np.sum(
        magnitude * np.asarray(1.0 - guess_mask), axis=axis, keepdims=True
    )
    denominator1 = (4.0 ** (-bits) / exponent_divisor) * np.sum(
        guess_mask, axis=axis, keepdims=True
    )
    denominator2 = np.sum(1.0 - guess_mask, axis=axis, keepdims=True)
    old_guess = guess
    guess = numerator / (denominator1 + denominator2)

  return guess


def get_tensor_quant_params(
    op_info: qtyping.OpInfo,
    tensor_quant_config: qtyping.TensorQuantizationConfig,
    tensor_content: Optional[np.ndarray] = None,
    tensor_qsv: Optional[dict[str, Any]] = None,
) -> qtyping.UniformQuantParams:
  """Returns the quantization parameters for a tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_quant_config: The quantization config for the tensor.
    tensor_content: The content of the tensor. When None, it means the tensor is
      not a weight tensor (e.g. static quantization) so we fallback to using
      naive_min_max_quantize.
    tensor_qsv: A dictionary containing the min/max of the tensor.

  Raises:
    ValueError: If the blockwise quantization is requested.
    ValueError: If the asymmetric quantization is requested.
    ValueError: `tensor_qsv` must contain min/max values, or `tensor_content`
      must be provided so that they can be inferred.
  """
  # Fallback to naive_min_max_quantize.py for non-weight tensors.
  if tensor_content is None:
    return naive_min_max_quantize.get_tensor_quant_params(
        op_info, tensor_quant_config, tensor_content, tensor_qsv
    )

  if not tensor_quant_config.symmetric:
    raise ValueError(
        f"Unsupported symmetry: {tensor_quant_config.symmetric}. OCTAV"
        " supports symmetric quantization only for now."
    )

  if not tensor_qsv:
    # We need min/max to calculate quantization parameters, which
    # should be collected during the calibration process. However,
    # weight-only and DRQ do not require calibration, thus it is
    # possible that this information is missing here. In that case we
    # collect min/max on the spot.
    tensor_min_max = common_quantize.init_tensor_min_max(
        tensor_content,
        op_info,
    )
  else:
    tensor_min_max = tensor_qsv

  if "min" not in tensor_min_max or "max" not in tensor_min_max:
    raise ValueError(
        "min and max must be provided to produce tensor quantization"
        " parameters. Check if the correct calibration results are passed into"
        " the ParamsGenerator."
    )

  quantized_dim = common_utils.get_weight_quantized_dim(
      op_info, tensor_content, tensor_quant_config.granularity
  )
  if uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity):
    reshaped_data, reduce_dims = (
        uniform_quantize_tensor.reshape_data_for_blockwise(
            tensor_content,
            op_info.op_name,
            tensor_quant_config.granularity,
        )
    )
  else:
    reshaped_data = tensor_content
    reduce_dims = common_utils.get_reduce_dims(
        quantized_dim, tensor_content.shape
    )
  clipping_constants = _guess_clipping_with_octav(
      reshaped_data,
      tensor_quant_config.num_bits,
      reduce_dims,
      max_iterations=10,
      exponent_divisor=3.0 if tensor_quant_config.symmetric else 12.0,
  )
  # We created a new dimension in order to reduce properly for blockwise
  # quantization, so we need to reshape the clipping constants back to the
  # min/max shape for the next step.
  if uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity):
    clipping_constants = clipping_constants.reshape(tensor_min_max["min"].shape)

  zp, scale = uniform_quantize_tensor.tensor_zp_scale_from_min_max(
      tensor_min_max["min"],
      tensor_min_max["max"],
      tensor_quant_config.num_bits,
      tensor_quant_config.symmetric,
      tensor_quant_config.granularity,
      clipping_constants,
  )

  quant_params = qtyping.UniformQuantParams(
      scale=scale,
      zero_point=zp,
      num_bits=tensor_quant_config.num_bits,
      symmetric=tensor_quant_config.symmetric,
      quantized_dimension=quantized_dim,
      block_size=uniform_quantize_tensor.extract_block_size_from_granularity(
          tensor_quant_config.granularity
      ),
  )

  quantized_vars = uniform_quantize_tensor.uniform_quantize(
      tensor_content,
      quant_params,
      is_blockwise_quant=uniform_quantize_tensor.is_blockwise(
          tensor_quant_config.granularity
      ),
  )

  return dataclasses.replace(quant_params, quantized_data=quantized_vars)



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with dynamic_update_slice."""

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  data = []
  for _ in range(num_samples):
    data.append({
        'input': _RNG.uniform(size=(8,)).astype(np.float32),
        'update': _RNG.uniform(size=(4,)).astype(np.float32),
        'indices': np.array([2], dtype=np.int32),
    })
  return data


def _get_calibration_data(num_samples: int = 8):
  calibration_samples = _get_dummy_data(num_samples)
  calibration_data = {'dynamic_update_slice': calibration_samples}
  return calibration_data


def _get_test_data(num_samples: int = 8):
  return _get_calibration_data(num_samples)


class DynamicUpdateSliceTest(parameterized.TestCase):

  def _custom_setup(self, test_model_file):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        f'../models/{test_model_file}'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  # DynamicUpdateSlice only currently supports 1-bit/8-bit/32-bit/64-bit
  # integer or float type.
  @parameterized.parameters(
      ('../../recipes/default_a8w8_recipe.json', 9),  # int8.
  )
  def test_dynamic_update_slice_model_full_integer(
      self, recipe_path, tensor_type
  ):
    self._custom_setup('dynamic_update_slice.tflite')
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quantization_result = self._quantizer.quantize(calibration_result)

    # Check input/output tensor type.
    quantized_model = tfl_flatbuffer_utils.read_model(
        quantization_result.quantized_model
    )
    self.assertLen(quantized_model.subgraphs, 1)
    subgraph = quantized_model.subgraphs[0]
    subgraph_tensors = subgraph.tensors
    self.assertLen(subgraph.inputs, 3)
    input_tensor = subgraph_tensors[subgraph.inputs[0]]
    update_tensor = subgraph_tensors[subgraph.inputs[1]]
    indices_tensor = subgraph_tensors[subgraph.inputs[2]]
    output_tensor = subgraph_tensors[subgraph.outputs[0]]
    # See schema_py_generated.py for type code.
    self.assertEqual(input_tensor.type, tensor_type)
    self.assertEqual(update_tensor.type, tensor_type)
    self.assertEqual(indices_tensor.type, 2)  # int32.
    self.assertEqual(output_tensor.type, tensor_type)

    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data(num_samples=1)
    )
    self._check_comparison_result(
        comparison_result,
        output_tolerance=1e-4,
    )

  # TODO: b/345503484 - Check weight tensor type of the quantized model.
  def _check_comparison_result(
      self,
      comparison_result,
      output_tolerance,
  ):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparison_result = comparison_result.get_all_tensor_results()
    # Check final output.
    output_mse = comparison_result['Identity_1']
    self.assertLess(output_mse, output_tolerance)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utils for handling operators with quantization constraints."""

from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_litert import schema_py_generated as schema_fb  # pylint: disable=g-direct-tensorflow-import


_OpQuantConstraint = common_utils.OpQuantConstraint


def get_constrained_op_list(
    quant_constraint: _OpQuantConstraint, verbose: bool = False
) -> list[str]:
  """Constructs and returns a list of constrained operators.

  This is achieved by invoking all materialization functions and extracting
  the constraint argument, using monkey patching to redirect logic to wrapper
  functions.

  Args:
    quant_constraint: The quantization constraint to filter operators by.
    verbose: Flag to enable verbose output.

  Returns:
    A list containing operators with the specified constraint.
  """
  constrained_ops = []

  def materialize_standard_op_wrapper(
      op_info: qtyping.OpInfo,
      *_args,
      constraint: _OpQuantConstraint = _OpQuantConstraint.NO_CONSTRAIN,
      **_kwargs,
  ) -> list[qtyping.TensorTransformationParams]:
    if constraint == quant_constraint:
      constrained_ops.append(op_info.op_name)
    # Return dummy values to avoid exceptions.
    dummy_value = [qtyping.TensorTransformationParams("")] * 2
    return dummy_value

  # Dummy implementation of the `_are_weights_too_small` function to support
  # `materialize_standard_op_wrapper` above.
  def are_weights_too_small_wrapper(*_args, **_kwargs) -> bool:
    return False

  # Dummy implementation of the `_materialize_bias_for_fc_conv_ops` function to
  # support `materialize_standard_op_wrapper` above.
  def materialize_bias_for_fc_conv_ops_wrapper(*_args, **_kwargs):
    return

  # Do monkey patch to intercept the `materialize_standard_op` function to
  # support `materialize_standard_op_wrapper` above.
  original_materialize_standard_op = common_utils.materialize_standard_op
  original_are_weights_too_small = common_quantize._are_weights_too_small  # pylint: disable=protected-access
  original_materialize_bias_for_fc_conv_ops = (
      common_quantize._materialize_bias_for_fc_conv_ops  # pylint: disable=protected-access
  )
  common_utils.materialize_standard_op = materialize_standard_op_wrapper
  common_quantize._are_weights_too_small = are_weights_too_small_wrapper  # pylint: disable=protected-access
  common_quantize._materialize_bias_for_fc_conv_ops = (  # pylint: disable=protected-access
      materialize_bias_for_fc_conv_ops_wrapper
  )
  minmax_func_dict = algorithm_manager.MIN_MAX_OP_NAME_MATERIALIZE_FUNC_DICT
  tensor_quant_params_cache = common_utils.TensorQuantParamsCache()

  # Loop over all available materialization functions to build up a list of
  # ops with the given constraint.
  for op, materialize_fn in minmax_func_dict.items():
    # Create a dummy op info to trigger the materialization.
    mock_op = schema_fb.OperatorT()
    mock_op.inputs = [0]
    mock_op.outputs = [0]
    op_info = qtyping.OpInfo(
        op=mock_op,
        op_name=op,
        subgraph_op_index=0,
        op_quant_config=qtyping.OpQuantizationConfig(),
    )
    materialize_fn(
        get_tensor_quant_params_fn=None,
        op_info=op_info,
        graph_info=None,
        tensor_name_to_qsv=None,
        tensor_quant_params_cache=tensor_quant_params_cache,
    )

  if verbose:
    print(f"  {quant_constraint} op list: {constrained_ops}")

  # Restore the original functions.
  common_utils.materialize_standard_op = original_materialize_standard_op
  common_quantize._are_weights_too_small = original_are_weights_too_small  # pylint: disable=protected-access
  common_quantize._materialize_bias_for_fc_conv_ops = (  # pylint: disable=protected-access
      original_materialize_bias_for_fc_conv_ops
  )
  return constrained_ops



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import octav
from ai_edge_quantizer.algorithms.uniform_quantize.op_architecture_tests import test_utils as op_test_utils
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

_TFLOpName = qtyping.TFLOperationName
_ComputePrecision = qtyping.ComputePrecision
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_QuantTransformation = qtyping.QuantTransformation
_OpTestInfo = op_test_utils.OpTestInfo


_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(
    "../../../tests/models"
)


class InputOutputTest(op_test_utils.BaseQuantizeTest):

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / "single_transpose.tflite"
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    self._model_qsv = {
        "serving_default_input_2:0": {
            "min": np.array([[-10]]),
            "max": np.array([[8]]),
        },
        "PartitionedCall:0": {"min": np.array([[10]]), "max": np.array([[88]])},
    }
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._input_op = qtyping.IOOperator(
        inputs=[],
        outputs=self._test_model.subgraphs[0].inputs,
        op_key=_TFLOpName.INPUT,
    )
    self._output_op = qtyping.IOOperator(
        inputs=self._test_model.subgraphs[0].outputs,
        outputs=[],
        op_key=_TFLOpName.OUTPUT,
    )

  @parameterized.product(
      weight_num_bits=[4, 8],
      symmetric_weight=[True, False],
      granularity=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_materialize_input_float(
      self, weight_num_bits, symmetric_weight, granularity
  ):
    weight_config = _TensorQuantConfig(
        num_bits=weight_num_bits,
        symmetric=symmetric_weight,
        granularity=granularity,
    )
    op_quant_config = qtyping.OpQuantizationConfig(
        activation_tensor_config=None,
        weight_tensor_config=weight_config,
        compute_precision=_ComputePrecision.FLOAT,
        explicit_dequantize=True,
    )
    op_info = qtyping.OpInfo(
        op=self._input_op,
        op_name=qtyping.TFLOperationName.INPUT,
        subgraph_op_index=-1,  # Virtual op, no real id.
        op_quant_config=op_quant_config,
    )
    quantization_params = common_quantize.materialize_input(
        naive_min_max_quantize.get_tensor_quant_params,
        op_info,
        self._graph_info,
        self._model_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    # Only one input tensor for the test model.
    self.assertLen(quantization_params, 1)
    quant_param = quantization_params[0]
    self.assertEqual(quant_param.tensor_name, "serving_default_input_2:0")
    # Produced by the virtual Input op.
    self.assertIsNotNone(quant_param.producer)
    producer_info = quant_param.producer
    self.assertEqual(producer_info.subgraph_op_id, -1)
    self.assertListEqual(
        producer_info.transformations, [qtyping.QuantTransformation.NO_QUANTIZE]
    )
    self.assertIsNone(producer_info.parameters)

  @parameterized.product(
      get_tensor_quant_params_func=(
          naive_min_max_quantize.get_tensor_quant_params,
          octav.get_tensor_quant_params,
      ),
      weight_num_bits=[4, 8],
      symmetric_weight=[True, False],
      granularity=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_materialize_output_float(
      self,
      get_tensor_quant_params_func,
      weight_num_bits,
      symmetric_weight,
      granularity,
  ):
    weight_config = _TensorQuantConfig(
        num_bits=weight_num_bits,
        symmetric=symmetric_weight,
        granularity=granularity,
    )
    op_quant_config = qtyping.OpQuantizationConfig(
        activation_tensor_config=None,
        weight_tensor_config=weight_config,
        compute_precision=_ComputePrecision.FLOAT,
        explicit_dequantize=True,
    )
    op_info = qtyping.OpInfo(
        op=self._output_op,
        op_name=qtyping.TFLOperationName.OUTPUT,
        subgraph_op_index=-1,  # Virtual op, no real id.
        op_quant_config=op_quant_config,
    )
    quantization_params = common_quantize.materialize_output(
        get_tensor_quant_params_func,
        op_info,
        self._graph_info,
        self._model_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    # Only one output tensor for the test model.
    self.assertLen(quantization_params, 1)
    quant_param = quantization_params[0]
    self.assertEqual(quant_param.tensor_name, "PartitionedCall:0")
    # Consumed by the virtual Output op.
    self.assertLen(quant_param.consumers, 1)
    consumer_info = quant_param.consumers[0]
    self.assertEqual(consumer_info.subgraph_op_id, -1)
    self.assertListEqual(
        consumer_info.transformations, [qtyping.QuantTransformation.NO_QUANTIZE]
    )
    self.assertIsNone(consumer_info.parameters)

  @parameterized.product(
      get_tensor_quant_params_func=(
          naive_min_max_quantize.get_tensor_quant_params,
          octav.get_tensor_quant_params,
      ),
      act_num_bits=[8, 16],
      weight_num_bits=[4, 8],
      granularity=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_materialize_input_integer(
      self,
      get_tensor_quant_params_func,
      act_num_bits,
      weight_num_bits,
      granularity,
  ):
    activation_config = _TensorQuantConfig(
        num_bits=act_num_bits,
        symmetric=True,
        granularity=qtyping.QuantGranularity.TENSORWISE,
    )
    weight_config = _TensorQuantConfig(
        num_bits=weight_num_bits,
        symmetric=True,
        granularity=granularity,
    )
    op_quant_config = qtyping.OpQuantizationConfig(
        activation_tensor_config=activation_config,
        weight_tensor_config=weight_config,
        compute_precision=_ComputePrecision.INTEGER,
    )
    op_info = qtyping.OpInfo(
        op=self._input_op,
        op_name=qtyping.TFLOperationName.INPUT,
        subgraph_op_index=-1,  # Virtual op, no real id.
        op_quant_config=op_quant_config,
    )
    quantization_params = common_quantize.materialize_input(
        get_tensor_quant_params_func,
        op_info,
        self._graph_info,
        self._model_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    # Only one input tensor for the test model.
    self.assertLen(quantization_params, 1)
    quant_param = quantization_params[0]
    self.assertEqual(quant_param.tensor_name, "serving_default_input_2:0")
    # Produced by the virtual Input op.
    self.assertIsNotNone(quant_param.producer)
    producer_info = quant_param.producer
    self.assertEqual(producer_info.subgraph_op_id, -1)
    self.assertListEqual(
        producer_info.transformations,
        [qtyping.QuantTransformation.ADD_DEQUANTIZE],
    )
    self.assertIsNotNone(producer_info.parameters)
    quant_params = producer_info.parameters
    self.assertEqual(quant_params.num_bits, act_num_bits)

  @parameterized.product(
      get_tensor_quant_params_func=(
          naive_min_max_quantize.get_tensor_quant_params,
          octav.get_tensor_quant_params,
      ),
      act_num_bits=[8, 16],
      weight_num_bits=[4, 8],
      granularity=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_materialize_output_integer(
      self,
      get_tensor_quant_params_func,
      act_num_bits,
      weight_num_bits,
      granularity,
  ):
    activation_config = _TensorQuantConfig(
        num_bits=act_num_bits,
        symmetric=True,
        granularity=qtyping.QuantGranularity.TENSORWISE,
    )
    weight_config = _TensorQuantConfig(
        num_bits=weight_num_bits,
        symmetric=True,
        granularity=granularity,
    )
    op_quant_config = qtyping.OpQuantizationConfig(
        activation_tensor_config=activation_config,
        weight_tensor_config=weight_config,
        compute_precision=_ComputePrecision.INTEGER,
    )
    op_info = qtyping.OpInfo(
        op=self._output_op,
        op_name=qtyping.TFLOperationName.OUTPUT,
        subgraph_op_index=-1,  # Virtual op, no real id.
        op_quant_config=op_quant_config,
    )
    quantization_params = common_quantize.materialize_output(
        get_tensor_quant_params_func,
        op_info,
        self._graph_info,
        self._model_qsv,
        tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
    )
    # Only one output tensor for the test model.
    self.assertLen(quantization_params, 1)
    quant_param = quantization_params[0]
    self.assertEqual(quant_param.tensor_name, "PartitionedCall:0")
    # Consumed by the virtual Output op.
    self.assertLen(quant_param.consumers, 1)
    consumer_info = quant_param.consumers[0]
    self.assertEqual(consumer_info.subgraph_op_id, -1)
    self.assertListEqual(
        consumer_info.transformations,
        [qtyping.QuantTransformation.ADD_QUANTIZE],
    )
    self.assertIsNotNone(consumer_info.parameters)
    quant_params = consumer_info.parameters
    self.assertEqual(quant_params.num_bits, act_num_bits)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Test insertion of the Hadamard rotation custom op."""

import pathlib
import absl.testing.absltest as absltest
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import insert_hadamard_rotation
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import

_TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('..')


class InsertHadamardRotationFullyConnectedTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc_bias.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)
    self.params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.ones(1),
        zero_point=np.zeros(1),
        hadamard=qtyping.UniformQuantParams.HadamardRotationParams(
            random_binary_vector=np.ones(1),
            hadamard_size=2,
        ),
    )

  def test_raise_unsupported_qparams(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'uniform quantization' in str(err)
    ):
      insert_hadamard_rotation.insert_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=qtyping.NonLinearQuantParams(
                  num_bits=16, quantized_data=None
              ),
          )
      )

  def test_raise_missing_hadamard_data(self):
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'quantization params are not set' in str(err)
    ):
      insert_hadamard_rotation.insert_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=qtyping.UniformQuantParams(
                  num_bits=8,
                  quantized_dimension=None,
                  scale=np.ones(1),
                  zero_point=np.zeros(1),
              ),
          )
      )

  def test_raise_non_float32_tensor(self):
    self.model.subgraphs[0].tensors[
        0
    ].type = schema_py_generated.TensorType.INT32
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: 'float32 tensors' in str(err)
    ):
      insert_hadamard_rotation.insert_hadamard_rotation(
          transformation_utils.TransformationInput(
              tensor_id=0,
              op_codes=self.model.operatorCodes,
              buffers=self.model.buffers,
              subgraph=self.model.subgraphs[0],
              producer=-1,
              consumers=[-1],
              quant_params=self.params,
          ),
      )

  def test_insert_single_custom_op(self):
    # Insert aeq.hadamard_rotation before fully_connected
    info = insert_hadamard_rotation.insert_hadamard_rotation(
        transformation_utils.TransformationInput(
            tensor_id=0,
            op_codes=self.model.operatorCodes,
            buffers=self.model.buffers,
            subgraph=self.model.subgraphs[0],
            producer=-1,
            consumers=[-1],
            quant_params=self.params,
        )
    )
    subgraph = self.model.subgraphs[0]
    self.assertEqual(info.op_id, 0)
    self.assertEqual(info.num_ops_added, 1)
    # Model had 4 tensors, added 1.
    self.assertEqual(info.output_tensor_id, 4)
    self.assertLen(subgraph.tensors, 5)
    # Model had 1 op, added a new one.
    self.assertLen(self.model.operatorCodes, 2)
    self.assertEqual(
        self.model.operatorCodes[1].builtinCode,
        schema_py_generated.BuiltinOperator.CUSTOM,
    )
    # First op is now the custom op, precedes fully_connected.
    self.assertEqual(
        self.model.operatorCodes[subgraph.operators[0].opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.CUSTOM,
    )
    # Input to the custom op is graph input
    self.assertEqual(subgraph.operators[0].inputs[0], 0)
    # Input to the FC is the custom op output
    self.assertEqual(subgraph.operators[1].inputs[0], 4)


class InsertHadamardRotationEmbeddingLookupTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(_TEST_DATA_PREFIX_PATH) / 'tests/models/embedding_lookup.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)
    self.params = qtyping.UniformQuantParams(
        num_bits=8,
        quantized_dimension=None,
        scale=np.ones(1),
        zero_point=np.zeros(1),
        hadamard=qtyping.UniformQuantParams.HadamardRotationParams(
            random_binary_vector=np.ones(1),
            hadamard_size=2,
        ),
    )

  def test_insert_single_custom_op(self):
    # Insert aeq.hadamard_rotation after embedding_lookup
    info = insert_hadamard_rotation.insert_hadamard_rotation(
        transformation_utils.TransformationInput(
            tensor_id=2,
            op_codes=self.model.operatorCodes,
            buffers=self.model.buffers,
            subgraph=self.model.subgraphs[0],
            producer=0,
            consumers=[-1],
            quant_params=self.params,
        )
    )
    subgraph = self.model.subgraphs[0]
    self.assertEqual(info.op_id, 1)
    self.assertEqual(info.num_ops_added, 1)
    # Model had 3 tensors, added 1.
    self.assertEqual(info.output_tensor_id, 3)
    self.assertLen(subgraph.tensors, 4)
    # Model had 1 op, added a new one.
    self.assertLen(self.model.operatorCodes, 2)
    self.assertEqual(
        self.model.operatorCodes[1].builtinCode,
        schema_py_generated.BuiltinOperator.CUSTOM,
    )
    # Second op is now the custom op, after embedding_lookup.
    self.assertEqual(
        self.model.operatorCodes[subgraph.operators[1].opcodeIndex].builtinCode,
        schema_py_generated.BuiltinOperator.CUSTOM,
    )
    # Input to the custom op is embedding's output
    self.assertEqual(subgraph.operators[1].inputs[0], 2)
    # Custom op's output is the new tensor
    self.assertEqual(subgraph.operators[1].outputs[0], 3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import collections
from absl.testing import absltest
from absl.testing import parameterized
from ai_edge_quantizer import default_policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.utils import common_utils

_ComputePrecision = qtyping.ComputePrecision
_QuantTransformation = qtyping.QuantTransformation
_TFLOpName = qtyping.TFLOperationName
_OpQuantConfig = qtyping.OpQuantizationConfig
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_DEFAULT_CONFIG_CHECK_POLICY = default_policy.DEFAULT_CONFIG_CHECK_POLICY


# TODO: b/335008966 - increase test coverage.
class MinMaxQuantizeUtilsTest(parameterized.TestCase):

  @parameterized.product(
      test_case=[
          # Tuple holds computation precision, whether to use SRQ and whether
          # to use explicit dequantize.
          (_ComputePrecision.FLOAT, False, True),  # WEIGHT_ONLY.
          (_ComputePrecision.INTEGER, False, False),  # DRQ.
          (_ComputePrecision.INTEGER, True, False),  # SRQ.
      ],
      is_inbounding_tensor=[True, False],
      is_constant=[True, False],
  )
  def test_get_tensor_transformations(
      self, test_case, is_inbounding_tensor, is_constant
  ):
    compute_precision, is_srq, explicit_dequantize = test_case
    weight_tensor_config = _TensorQuantConfig(num_bits=8)
    op_quant_config = qtyping.OpQuantizationConfig(
        activation_tensor_config=weight_tensor_config if is_srq else None,
        compute_precision=compute_precision,
        explicit_dequantize=explicit_dequantize,
    )
    transformations = common_utils.get_tensor_transformations(
        op_quant_config, is_inbounding_tensor, is_constant
    )
    # Check if WEIGHT_ONLY.
    if (
        compute_precision == _ComputePrecision.FLOAT
        and op_quant_config.explicit_dequantize
    ):
      if is_inbounding_tensor and is_constant:
        self.assertSequenceEqual(
            transformations,
            [
                _QuantTransformation.ADD_DEQUANTIZE,
            ],
        )
      else:
        self.assertSequenceEqual(
            transformations,
            [_QuantTransformation.NO_QUANTIZE],
        )

    # Check if DRQ.
    if compute_precision == _ComputePrecision.INTEGER and not is_srq:
      if is_inbounding_tensor and is_constant:
        self.assertSequenceEqual(
            transformations, [_QuantTransformation.QUANTIZE_TENSOR]
        )
      else:
        self.assertSequenceEqual(
            transformations,
            [_QuantTransformation.NO_QUANTIZE],
        )

    # Check if SRQ.
    if compute_precision == _ComputePrecision.INTEGER and is_srq:
      if is_inbounding_tensor:
        if is_constant:
          self.assertSequenceEqual(
              transformations, [_QuantTransformation.QUANTIZE_TENSOR]
          )
        else:
          self.assertSequenceEqual(
              transformations, [_QuantTransformation.ADD_QUANTIZE]
          )
      else:
        self.assertSequenceEqual(
            transformations, [_QuantTransformation.ADD_DEQUANTIZE]
        )

  @parameterized.parameters((_TFLOpName.FULLY_CONNECTED), (_TFLOpName.CONV_2D))
  def test_check_weight_only_config_succeeds(self, op_name):
    self.assertIn(op_name, _DEFAULT_CONFIG_CHECK_POLICY.keys())

  @parameterized.parameters((_TFLOpName.RESHAPE), (_TFLOpName.AVERAGE_POOL_2D))
  def test_check_weight_only_config_raises_when_invalid_config(self, op_name):
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
        ),
        compute_precision=_ComputePrecision.FLOAT,
    )
    error_message = (
        f"Quantization config for op: {op_name} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  @parameterized.product(
      op_name=(_TFLOpName.FULLY_CONNECTED, _TFLOpName.CONV_2D),
      weight_num_bits=(4, 8),
      granularity=(
          qtyping.QuantGranularity.TENSORWISE,
          qtyping.QuantGranularity.CHANNELWISE,
      ),
  )
  def test_check_drq_config_succeeds(
      self, op_name, weight_num_bits, granularity
  ):
    # TODO: b/353365054 - Remove this check after int4 DRQ is supported for
    # conv2d.
    if op_name == _TFLOpName.CONV_2D and weight_num_bits == 4:
      return
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_num_bits,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
    )
    common_utils.check_if_valid_op_config(
        op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
    )

  @parameterized.parameters((_TFLOpName.RESHAPE), (_TFLOpName.AVERAGE_POOL_2D))
  def test_check_drq_config_unsupported_op_raise_error(self, op_name):
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
    )
    error_message = (
        f"Quantization config for op: {op_name} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  @parameterized.parameters((_TFLOpName.FULLY_CONNECTED), (_TFLOpName.CONV_2D))
  def test_check_drq_config_wrong_bits_raise_error(self, op_name):
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=2,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
    )
    error_message = (
        f"Quantization config for op: {op_name} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  @parameterized.parameters((_TFLOpName.FULLY_CONNECTED), (_TFLOpName.CONV_2D))
  def test_check_drq_config_asymmetric_weights_raise_error(self, op_name):
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            symmetric=False,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
    )
    error_message = (
        f"Quantization config for op: {op_name} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  def test_check_drq_config_with_non_default_min_weight_elements_succeeds(self):
    op_quant_config = _OpQuantConfig(
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
        min_weight_elements=100,
    )
    common_utils.check_if_valid_op_config(
        _TFLOpName.CONV_2D, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
    )

  @parameterized.product(
      op_name=(_TFLOpName.FULLY_CONNECTED, _TFLOpName.CONV_2D),
      act_num_bits=(8, 16),
      weight_num_bits=(4, 8),
      granularity=(
          qtyping.QuantGranularity.TENSORWISE,
          qtyping.QuantGranularity.CHANNELWISE,
      ),
      symmetric_act=(True, False),
  )
  def test_check_srq_config_succeeds(
      self,
      op_name,
      act_num_bits,
      weight_num_bits,
      granularity,
      symmetric_act,
  ):
    # Asym int16 activation is not supported.
    if not symmetric_act and act_num_bits == 16:
      return
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=act_num_bits, symmetric=symmetric_act
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=weight_num_bits,
            granularity=granularity,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    common_utils.check_if_valid_op_config(
        op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
    )

  def test_check_srq_config_unsupported_op_raise_error(self):
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(num_bits=8, symmetric=True),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    error_message = (
        f"Unsupported op for {op_quant_config.compute_precision}:"
        f" {_TFLOpName.CUSTOM_OP}"
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          _TFLOpName.CUSTOM_OP, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  def test_check_srq_config_wrong_act_bits_config_raise_error(self):
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=14, symmetric=True
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    error_message = (
        f"Quantization config for op: {_TFLOpName.FULLY_CONNECTED} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          _TFLOpName.FULLY_CONNECTED,
          op_quant_config,
          _DEFAULT_CONFIG_CHECK_POLICY,
      )

  def test_check_srq_config_asym_int16_act_raise_error(self):
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=16, symmetric=False
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    error_message = (
        f"Quantization config for op: {_TFLOpName.FULLY_CONNECTED} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          _TFLOpName.FULLY_CONNECTED,
          op_quant_config,
          _DEFAULT_CONFIG_CHECK_POLICY,
      )

  def test_check_srq_config_wrong_weight_bits_raise_error(self):
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(
            num_bits=16, symmetric=True
        ),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=2,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    error_message = (
        f"Quantization config for op: {_TFLOpName.FULLY_CONNECTED} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          _TFLOpName.FULLY_CONNECTED,
          op_quant_config,
          _DEFAULT_CONFIG_CHECK_POLICY,
      )

  def test_check_srq_config_asym_weight_raise_error(self):
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=_TensorQuantConfig(num_bits=8, symmetric=True),
        weight_tensor_config=_TensorQuantConfig(
            num_bits=8,
            symmetric=False,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=_ComputePrecision.INTEGER,  # SRQ.
    )
    error_message = (
        f"Quantization config for op: {_TFLOpName.FULLY_CONNECTED} with config:"
        f" {op_quant_config} was not found in the policy."
    )
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      common_utils.check_if_valid_op_config(
          _TFLOpName.FULLY_CONNECTED,
          op_quant_config,
          _DEFAULT_CONFIG_CHECK_POLICY,
      )

  @parameterized.product(
      op_name=[
          _TFLOpName.FULLY_CONNECTED,
          _TFLOpName.CONV_2D,
      ],
      activation_tensor_config=[
          None,
          _TensorQuantConfig(num_bits=8, symmetric=False),
          _TensorQuantConfig(num_bits=16, symmetric=True),
      ],
      compute_precision=[
          _ComputePrecision.FLOAT,
          _ComputePrecision.INTEGER,
      ],
  )
  def test_check_supported_int4_config_succeeds(
      self, op_name, activation_tensor_config, compute_precision
  ):
    # Exclude invalid SRQ config.
    if (
        activation_tensor_config is not None
        and compute_precision != _ComputePrecision.INTEGER
    ) or (
        activation_tensor_config is None
        and compute_precision == _ComputePrecision.FLOAT
    ):
      return
    # TODO: b/353365054 - Remove this check after int4 DRQ is supported for
    # conv2d.
    if (
        # Check if DRQ and CONV_2D.
        compute_precision == _ComputePrecision.INTEGER
        and activation_tensor_config is None
        and op_name == _TFLOpName.CONV_2D
    ):
      return
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=activation_tensor_config,
        weight_tensor_config=_TensorQuantConfig(
            num_bits=4,
            symmetric=True,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=compute_precision,
    )
    # Raise error if the config is not supported.
    # Check if DRQ.
    if (
        compute_precision == _ComputePrecision.INTEGER
        and op_quant_config.activation_tensor_config is None
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )
    # Check if WEIGHT_ONLY.
    elif (
        compute_precision == _ComputePrecision.FLOAT
        and op_quant_config.explicit_dequantize
    ):
      self.assertIn(op_name, _DEFAULT_CONFIG_CHECK_POLICY.keys())
    # Check if SRQ.
    if (
        compute_precision == _ComputePrecision.INTEGER
        and op_quant_config.activation_tensor_config is not None
    ):
      common_utils.check_if_valid_op_config(
          op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
      )

  @parameterized.product(
      op_name=[_TFLOpName.BATCH_MATMUL],
      activation_tensor_config=[
          None,
          _TensorQuantConfig(num_bits=8, symmetric=False),
          _TensorQuantConfig(num_bits=16, symmetric=True),
      ],
      test_case=[
          # Tuple holds compute precision and whether to use drq.
          (_ComputePrecision.INTEGER, True),
          (_ComputePrecision.INTEGER, False),
      ],
  )
  def test_check_unsupported_int4_config_raise_error(
      self, op_name, activation_tensor_config, test_case
  ):
    compute_precision, is_drq = test_case
    # Exclude invalid SRQ config.
    if (activation_tensor_config is not None and is_drq) or (
        activation_tensor_config is None and not is_drq
    ):
      return
    op_quant_config = _OpQuantConfig(
        activation_tensor_config=activation_tensor_config,
        weight_tensor_config=_TensorQuantConfig(
            num_bits=4,
            symmetric=True,
            granularity=qtyping.QuantGranularity.CHANNELWISE,
        ),
        compute_precision=compute_precision,
    )

    with self.assertRaises(ValueError):
      if is_drq:
        common_utils.check_if_valid_op_config(
            op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
        )
      elif not is_drq:
        common_utils.check_if_valid_op_config(
            op_name, op_quant_config, _DEFAULT_CONFIG_CHECK_POLICY
        )

  def test_materialize_op_with_output_activation_constraint_fails_for_multiple_output_op(
      self,
  ):
    # Create a mock op with multiple outputs.
    MockOp = collections.namedtuple("MockOp", ["outputs"])
    mock_op_info = qtyping.OpInfo(
        op=MockOp(outputs=[1, 2]),
        op_name=_TFLOpName.SOFTMAX,
        subgraph_op_index=0,
        op_quant_config=_OpQuantConfig(),
    )

    with self.assertRaisesRegex(
        ValueError, "only supports ops with a single output tensor"
    ):
      common_utils.materialize_op_with_output_activation_constraint(
          op_info=mock_op_info,
          graph_info=qtyping.GraphInfo([], []),
          tensor_name_to_qsv={},
          output_activation_constraints={},
          get_tensor_quant_params_fn=lambda *args: [],
          tensor_quant_params_cache=common_utils.TensorQuantParamsCache(),
      )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Hadamard rotation decomposed pattern transformation."""

from flatbuffers import flexbuffers
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


def _to_flexbuffer(
    hadamard_size: int,
    random_binary_vector: list[np.int8],
) -> bytes:
  """Converts hadamard_size to flexbuffer."""
  fbb = flexbuffers.Builder()
  with fbb.Map():
    fbb.Int('hadamard_size', hadamard_size)
    fbb.VectorFromElements('random_binary_vector', random_binary_vector)
  return fbb.Finish()


def _update_embedding_lookup_consumers(
    transformation: transformation_utils.TransformationInput,
    new_tensor_id: int,
) -> bool:
  """Updates the consumers of the embedding lookup op to use the new tensor.

  Args:
    transformation: The transformation input to update the consumers of.
    new_tensor_id: The new tensor id to use as the input to the embedding lookup
      consumers.
  """
  for consumer in transformation.consumers:
    # If the consumer is a graph output and not an op, we can ignore it here
    # since the graph output will be updated later.
    if consumer == -1:
      continue
    consumer_op = transformation.subgraph.operators[consumer]
    # Find the input that was attached to the insertion point, and replace it
    # with the new tensor.
    for i in range(len(consumer_op.inputs)):
      if consumer_op.inputs[i] == transformation.tensor_id:
        consumer_op.inputs[i] = new_tensor_id


def _update_fully_connected_consumers(
    transformation: transformation_utils.TransformationInput,
    new_tensor_id: int,
) -> bool:
  """Updates the fully connected op(s) to use the new tensor.

  Since the new tensor is inserted to the fully_connected's input, we need to
  scan each consumer (in case of multiple fully_connected ops), and update
  the input tensor to the new tensor.

  Args:
    transformation: The transformation input to update the consumers of.
    new_tensor_id: The new tensor id to use as the input to the fully connected
      consumers.

  Returns:
    True if the fully connected op(s) were updated to use the new tensor.
  """
  updated = False
  for consumer in transformation.consumers:
    if (
        transformation_utils.get_schema_op_id(transformation, consumer)
        == schema_py_generated.BuiltinOperator.FULLY_CONNECTED
    ):
      transformation.subgraph.operators[consumer].inputs[0] = new_tensor_id
      updated = True
  return updated


def _make_hadamard_matrix(size: int):
  """Generates an unnormalized integer Hadamard matrix of the given size.

  Args:
    size: The size of the Hadamard matrix. Must be a power of 2. This represents
      a single dimension. E.g. if size is 4, then the Hadamard matrix is a 4x4
      matrix.

  Returns:
    The Hadamard matrix.

  Raises:
    ValueError: If the size is not a power of 2.
  """
  if size <= 0 or (size & (size - 1)) != 0:
    raise ValueError('Hadamard matrix size must be a power of 2. ')
  h = h2 = np.array([[1, 1], [1, -1]], dtype=np.int8)
  current_size = 2
  while current_size < size:
    h = np.kron(h, h2)
    current_size *= 2
  return h


def insert_decomposed_hadamard_rotation(
    transformation_input: transformation_utils.TransformationInput,
) -> qtyping.TransformationInfo:
  """Inserts a decomposed pattern of Hadamard rotation on this tensor.

  This function works for float32 tensors only. Instead of inserting a single
  custom op (aeq.hadamard_rotation), this inserts the mathematical equivalent
  expressed in built-in TFLite ops. The mathematical equivalent is:
    x' = reshape(x, (-1, hadamard_size))
    x' = x' @ H(hadamard_size)
    x' = reshape(x, x.shape)
  where H(n) is a Hadamard matrix of size n.

  Args:
    transformation_input: The transformation input to insert the ops on.

  Returns:
    The transformation info of the inserted ops.

  Raises:
    ValueError: If the transformation input is not a uniform quantization
    transformation.
    ValueError: If the Hadamard quantization params are not set.
    ValueError: If the tensor is not a float32 tensor.
    ValueError: If no supported ops were found as the tensor's producer or
    consumers.
  """
  if not isinstance(
      transformation_input.quant_params, qtyping.UniformQuantParams
  ):
    raise ValueError('Hadamard rotation supports uniform quantization only')

  if transformation_input.quant_params.hadamard is None:
    raise ValueError(
        'Hadamard rotation quantization params are not set but op insertion is'
        ' requested.'
    )

  tensor = transformation_input.subgraph.tensors[transformation_input.tensor_id]
  if tensor.type != schema_py_generated.TensorType.FLOAT32:
    raise ValueError(
        'The Hadamard rotation op supports float32 tensors only. Got'
        f' {tensor.type} tensor.'
    )

  # Insert x' = tfl.reshape to reshape x to (-1, hadamard_size)
  hadamard_size = transformation_input.quant_params.hadamard.hadamard_size
  tensor_size = np.prod(tensor.shape)
  num_hadamard_blocks = tensor_size // hadamard_size
  prerotate_shape = [num_hadamard_blocks, hadamard_size]
  prerotate_shape_tensor_id = transformation_utils.add_new_constant_tensor(
      tensor.name + b'_prerotate_shape',
      np.array(prerotate_shape, dtype=np.int32),
      schema_py_generated.TensorType.INT32,
      transformation_input.subgraph,
      transformation_input.buffers,
  )
  prerotate_reshape_output_tensor_id = (
      transformation_utils.add_new_activation_tensor(
          tensor.name + b'_prerotate_reshaped',
          prerotate_shape,
          schema_py_generated.TensorType.FLOAT32,
          transformation_input.subgraph,
      )
  )

  prerotate_reshape_op_code_idx = transformation_utils.add_op_code(
      schema_py_generated.BuiltinOperator.RESHAPE,
      transformation_input.op_codes,
      'RESHAPE',
  )
  prerorate_reshape_op = schema_py_generated.OperatorT()
  prerorate_reshape_op.opcodeIndex = prerotate_reshape_op_code_idx
  prerorate_reshape_op.inputs = [
      transformation_input.tensor_id,
      prerotate_shape_tensor_id,
  ]
  prerorate_reshape_op.outputs = [prerotate_reshape_output_tensor_id]

  # Generate hadamard_matrix(hadamard_size).
  # We quantize the Hadamard matrix to INT4 for better memory efficiency, but
  # for large models the memory overhead is not significant, and floating point
  # computation does seem to result in better accuracy.
  hadamard_matrix = _make_hadamard_matrix(hadamard_size)
  hadamard_matrix_tensor_id = transformation_utils.add_new_constant_tensor(
      tensor_name=tensor.name + b'_hadamard_matrix',
      data=transformation_utils.pack_data(
          bitwidth=4, flattened_data=hadamard_matrix.flatten()
      ),
      tensor_type=schema_py_generated.TensorType.INT4,
      subgraph=transformation_input.subgraph,
      buffers=transformation_input.buffers,
      tensor_shape=hadamard_matrix.shape,
      quantization=schema_py_generated.QuantizationParametersT(
          scale=np.array([1.0 / np.sqrt(hadamard_size)], dtype=np.float32),
          zeroPoint=[0],
      ),
  )

  # Insert x' = tfl.fully_connected(x', hadamard_matrix)
  fc_output_tensor_id = transformation_utils.add_new_activation_tensor(
      tensor.name + b'_rotated',
      prerotate_shape,
      schema_py_generated.TensorType.FLOAT32,
      transformation_input.subgraph,
  )

  fc_op_code_idx = transformation_utils.add_op_code(
      schema_py_generated.BuiltinOperator.FULLY_CONNECTED,
      transformation_input.op_codes,
      'FULLY_CONNECTED',
  )
  fc_op = schema_py_generated.OperatorT()
  fc_op.opcodeIndex = fc_op_code_idx
  fc_op.inputs = [prerotate_reshape_output_tensor_id, hadamard_matrix_tensor_id]
  fc_op.outputs = [fc_output_tensor_id]
  fc_options = schema_py_generated.FullyConnectedOptionsT()
  fc_options.fusedActivationFunction = (
      schema_py_generated.ActivationFunctionType.NONE
  )
  fc_op.builtinOptionsType = (
      schema_py_generated.BuiltinOptions.FullyConnectedOptions
  )
  fc_op.builtinOptions = fc_options

  # Insert x' = tfl.reshape(x', x.shape)
  post_reshape_op_code_idx = transformation_utils.add_op_code(
      schema_py_generated.BuiltinOperator.RESHAPE,
      transformation_input.op_codes,
      'RESHAPE',
  )
  post_reshape_op = schema_py_generated.OperatorT()
  post_reshape_op.opcodeIndex = post_reshape_op_code_idx
  post_reshape_shape_tensor_id = transformation_utils.add_new_constant_tensor(
      tensor.name + b'_postrotate_shape',
      np.array(tensor.shape, dtype=np.int32),
      schema_py_generated.TensorType.INT32,
      transformation_input.subgraph,
      transformation_input.buffers,
  )

  post_reshape_output_tensor_id = (
      transformation_utils.add_new_activation_tensor(
          tensor.name + b'_postrotate_reshaped',
          tensor.shape,
          schema_py_generated.TensorType.FLOAT32,
          transformation_input.subgraph,
      )
  )
  post_reshape_op.inputs = [
      fc_output_tensor_id,
      post_reshape_shape_tensor_id,
  ]
  post_reshape_op.outputs = [post_reshape_output_tensor_id]

  # Update the users of this tensor to use the new tensor.
  if (
      transformation_utils.get_producer_schema_op_id(transformation_input)
      == schema_py_generated.BuiltinOperator.EMBEDDING_LOOKUP
  ):
    _update_embedding_lookup_consumers(
        transformation_input, post_reshape_output_tensor_id
    )
  elif not _update_fully_connected_consumers(
      transformation_input, post_reshape_output_tensor_id
  ):
    raise ValueError(
        'The Hadamard rotation op supports embedding lookup and fully connected'
        ' ops only, but no such ops were found.'
    )

  # If the tensor is a graph output, we need to replace the tensor with the
  # new tensor.
  for i, output in enumerate(transformation_input.subgraph.outputs):
    if output == transformation_input.tensor_id:
      transformation_input.subgraph.outputs[i] = post_reshape_output_tensor_id

  # Find the actual insertion point. The insertion point should be after the
  # producer op and before the first consumer op. The max() operation ensures
  # that we're not using -1 as the insertion point.
  first_consumer_id = min(transformation_input.consumers)
  op_id = max(transformation_input.producer + 1, first_consumer_id)

  # Insert the new ops in the correct order.
  transformation_input.subgraph.operators.insert(op_id, prerorate_reshape_op)
  transformation_input.subgraph.operators.insert(op_id + 1, fc_op)
  transformation_input.subgraph.operators.insert(op_id + 2, post_reshape_op)

  return qtyping.TransformationInfo(
      op_id=op_id,
      num_ops_added=3,
      output_tensor_id=post_reshape_output_tensor_id,
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with depthwise conv2d."""

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  data = []
  for _ in range(num_samples):
    data.append(
        {'input_1': _RNG.uniform(size=(1, 64, 64, 3)).astype(np.float32)}
    )
  return data


def _get_calibration_data(num_samples: int = 128):
  calibration_samples = _get_dummy_data(num_samples)
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _get_test_data(num_samples: int = 8):
  return _get_calibration_data(num_samples)


class DepthwiseConv2dTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/single_depthwise_conv2d_bias.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.product(
      symmetric_weight=[True, False],
      channel_wise_weight=[
          qtyping.QuantGranularity.CHANNELWISE,
          qtyping.QuantGranularity.TENSORWISE,
      ],
  )
  def test_depthwise_conv2d_model_int8_weight_only(
      self, symmetric_weight, channel_wise_weight
  ):
    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=_OpName.DEPTHWISE_CONV_2D,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=8,
                symmetric=symmetric_weight,
                granularity=channel_wise_weight,
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY.
            explicit_dequantize=True,
        ),
    )
    quant_result = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(quant_result.quantized_model), 10000)

    comparion_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparion_result(
        comparion_result,
        weight_tolerance=1e-2 if channel_wise_weight else 1e-1,
        output_tolerance=1e-4 if channel_wise_weight else 1e-2,
    )

  @parameterized.parameters(
      '../../recipes/dynamic_legacy_wi8_afp32_recipe.json',
      '../../recipes/dynamic_wi8_afp32_recipe.json',
  )
  def test_depthwise_conv2d_model_int8_drq(self, recipe_path):
    full_recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(full_recipe_path)

    # Check model size.
    quant_result = self._quantizer.quantize()
    self.assertLess(len(quant_result.quantized_model), 10000)

    comparion_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparion_result(
        comparion_result,
        weight_tolerance=1e-2,
        output_tolerance=1e-4,
    )

  @parameterized.parameters(
      '../../recipes/default_a8w8_recipe.json',
      '../../recipes/default_a16w8_recipe.json',
  )
  def test_depthwise_conv2d_model_full_integer(self, recipe_path):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    quant_result = self._quantizer.quantize(calibration_result)
    # Check model size.
    self.assertLess(len(quant_result.quantized_model), 10000)

    comparion_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparion_result(
        comparion_result,
        weight_tolerance=1e-2,
        output_tolerance=1e-4,
    )

  def test_depthwise_conv2d_model_fp16_weight_only(self):
    self._quantizer.update_quantization_recipe(
        regex='.*',
        algorithm_key=quantizer.AlgorithmName.FLOAT_CASTING,
        operation_name=_OpName.DEPTHWISE_CONV_2D,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY
            explicit_dequantize=True,
        ),
    )
    quant_result = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(quant_result.quantized_model), 10000)

    comparion_result = self._quantizer.validate(error_metrics='mse')
    self._check_comparion_result(
        comparion_result,
        weight_tolerance=1e-5,
        output_tolerance=1e-5,
    )

  # TODO: b/345503484 - Check weight tensor type of the quantized model.
  def _check_comparion_result(
      self,
      comparion_result,
      weight_tolerance,
      output_tolerance,
  ):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparion_result = comparion_result.get_all_tensor_results()
    # Check weight tensors.
    weight_mse = comparion_result['sequential/depthwise_conv2d/depthwise']
    self.assertLess(weight_mse, weight_tolerance)
    bias_mse = comparion_result[
        'sequential/depthwise_conv2d/BiasAdd;sequential/depthwise_conv2d/depthwise;sequential/depthwise_conv2d/BiasAdd/ReadVariableOp'
    ]
    self.assertLess(bias_mse, weight_tolerance)
    # Check final output.
    output_mse = comparion_result['StatefulPartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
import numpy as np
import absl.testing.absltest as absltest
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import duplicate_tensor
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('..')


class DuplicateTensorTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/weight_sharing_fcs.tflite'
    )
    self.model = tfl_flatbuffer_utils.read_model(model_path)

  def _get_transformation_input(
      self,
      subgraph_idx: int,
      tensor_idx: int,
      consumers: list[int],
  ) -> transformation_utils.TransformationInput:
    return transformation_utils.TransformationInput(
        tensor_id=tensor_idx,
        buffers=self.model.buffers,
        consumers=consumers,
        # Dummy params below.
        op_codes=self.model.operatorCodes,
        subgraph=self.model.subgraphs[subgraph_idx],
        producer=-1,
        quant_params=qtyping.UniformQuantParams(
            num_bits=8,
            quantized_dimension=None,
            scale=np.ones(1),
            zero_point=np.zeros(1),
        ),
    )

  def test_constant_tensor_is_correctly_duplicated(self):
    # Duplicate the FC weight tensor in the second subgraph for the first FC.
    subgraph_idx = 1
    fc1_op_idx = 0
    prev_weight_tensor_idx = 1
    subgraph = self.model.subgraphs[subgraph_idx]
    weight_idx_in_op_inputs = list(subgraph.operators[fc1_op_idx].inputs).index(
        prev_weight_tensor_idx
    )
    prev_num_tensors = len(subgraph.tensors)
    prev_buffer_id = subgraph.tensors[prev_weight_tensor_idx].buffer
    prev_num_buffers = len(self.model.buffers)
    transformation_input = self._get_transformation_input(
        subgraph_idx, prev_weight_tensor_idx, consumers=[fc1_op_idx]
    )
    transformation_info = duplicate_tensor.duplicate_tensor(
        transformation_input
    )
    self.assertEqual(transformation_info.op_id, 0)
    self.assertEqual(transformation_info.num_ops_added, 0)
    # Check that a new tensor and buffer were added.
    self.assertLen(subgraph.tensors, prev_num_tensors + 1)
    self.assertLen(self.model.buffers, prev_num_buffers + 1)
    # Check that the duplicated tensor is the last tensor in the subgraph.
    weight_tensor_idx = transformation_info.output_tensor_id
    self.assertEqual(weight_tensor_idx, len(subgraph.tensors) - 1)
    # Compare tensors.
    original_tensor = subgraph.tensors[prev_weight_tensor_idx]
    original_tensor_name = tfl_flatbuffer_utils.get_tensor_name(original_tensor)
    duplicated_tensor = subgraph.tensors[weight_tensor_idx]
    self.assertEqual(
        duplicated_tensor.name,
        f'{original_tensor_name}_duplicated_{weight_tensor_idx}',
    )
    self.assertEqual(duplicated_tensor.type, original_tensor.type)
    self.assertTrue(np.all(duplicated_tensor.shape == original_tensor.shape))
    # Check that the new buffer is used by the duplicated tensor.
    new_buffer_id = len(self.model.buffers) - 1
    self.assertEqual(duplicated_tensor.buffer, new_buffer_id)
    # Check that the new buffer has the same data as the original one.
    self.assertTrue(
        np.all(
            np.frombuffer(
                self.model.buffers[new_buffer_id].data,
                dtype=np.float32,
            )
            == np.frombuffer(
                self.model.buffers[prev_buffer_id].data,
                dtype=np.float32,
            )
        )
    )
    # Check that first FC input tensor id was updated.
    self.assertEqual(
        subgraph.operators[fc1_op_idx].inputs[weight_idx_in_op_inputs],
        weight_tensor_idx,
    )

  def test_duplicate_tensor_raises_error_when_tensor_is_not_constant(self):
    # Duplicate the FC input tensor in the second subgraph.
    subgraph_idx = 1
    input_tensor_idx = 0
    transformation_input = self._get_transformation_input(
        subgraph_idx, input_tensor_idx, consumers=[0]
    )
    with self.assertRaisesRegex(
        ValueError,
        'Duplicate Tensor transformation supports only constant tensors.',
    ):
      duplicate_tensor.duplicate_tensor(transformation_input)

if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with tanh op."""

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_OpExecutionMode = qtyping.OpExecutionMode
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  samples = []
  for _ in range(num_samples):
    samples.append({'input_1': _RNG.uniform(size=(1, 10)).astype(np.float32)})

  data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: samples,
  }
  return data


def _get_calibration_data(num_samples: int = 128):
  return _get_dummy_data(num_samples)


def _get_test_data(num_samples: int = 8):
  return _get_dummy_data(num_samples)


class LogisticTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/single_fc_bias_logistic.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.parameters(
      '../../recipes/default_a8w8_recipe.json',
      '../../recipes/default_a16w8_recipe.json',
  )
  def test_logistic_model_full_integer(self, recipe_path):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    _ = self._quantizer.quantize(calibration_result)

    comparison_result = self._quantizer.validate(
        error_metrics='mse', test_data=_get_test_data()
    )
    self._check_comparison_result(comparison_result, output_tolerance=1e-3)

  # TODO: b/345503484 - Check weight tensor type of the quantized model.
  def _check_comparison_result(self, comparison_result, output_tolerance):
    # TODO: b/357959309 - Use comparison result directly for testing.
    comparison_result = comparison_result.get_all_tensor_results()
    output_mse = comparison_result['StatefulPartitionedCall:0']
    self.assertLess(output_mse, output_tolerance)


if __name__ == '__main__':
  absltest.main()
